import matplotlib.pyplot as plt
import numpy as np
from PIL import Image
import sys
from typing import Optional

from pyRDDLGym.Core.Compiler.RDDLModel import PlanningModel
from pyRDDLGym import Visualizer
from pyRDDLGym.Visualizer.StateViz import StateViz


class HVACVisualizer(StateViz):

    def __init__(self, model: PlanningModel,
                 grid_size: Optional[int] = [50, 50],
                 resolution: Optional[int] = [500, 500]) -> None:
        self._model = model
        self._states = model.groundstates()
        self._nonfluents = model.groundnonfluents()
        self._objects = model.objects
        self._grid_size = grid_size
        self._resolution = resolution
        self._interval = 15

        self._asset_path = "/".join(Visualizer.__file__.split("/")[:-1])

        self._object_layout = None
        self._canvas_info = None
        self._render_frame = None
        self._data = None
        self._fig = None
        self._ax = None

    def build_object_layout(self) -> dict:

        # print(self._nonfluents)

        zone_vol = {o: None for o in self._objects['zone']}
        heater_vol = {o: None for o in self._objects['heater']}
        adj_zone = {o: [] for o in self._objects['zone']}
        adj_heater = {o: [] for o in self._objects['heater']}
        temp_zone = {o: None for o in self._objects['zone']}
        temp_heater = {o: None for o in self._objects['heater']}

        # add none-fluents
        for k, v in self._nonfluents.items():
            var, objects = self._model.parse(k)
            if var == 'ZONE_VOL':
                zone_vol[objects[0]] = v
            elif var == 'HEATER-VOL':
                heater_vol[objects[0]] = v
            elif var == 'ADJ_ZONES':
                if v == True:
                    adj_zone[objects[0]].append(objects[1])
            elif var == 'ADJ_HEATER':
                if v == True:
                    adj_heater[objects[0]].append(objects[1])


            # if 'ZONE-VOL_' in k:
            #     point = k.split('___')[1]
            #     zone_vol[point] = v
            # elif 'HEATER-VOL_' in k:
            #     point = k.split('___')[1]
            #     heater_vol[point] = v
            # elif 'ADJ-ZONES_' in k:
            #     if v == True:
            #         point = k.split('___')[1]
            #         v = point.split('__')
            #         adj_zone[v[0]].append(v[1])
            # elif 'ADJ-HEATER_' in k:
            #     if v == True:
            #         point = k.split('___')[1]
            #         v = point.split('__')
            #         adj_heater[v[0]].append(v[1])

        # add states
        for k, v in self._states.items():
            var, objects = self._model.parse(k)
            if var == 'temp-zone':
                temp_zone[objects[0]] = v
            elif var == 'temp-heater':
                temp_heater[objects[0]] = v
            # if 'temp-zone_' in k:
            #     point = k.split('___')[1]
            #     temp_zone[point] = v
            # elif 'temp-heater_' in k:
            #     point = k.split('___')[1]
            #     temp_heater[point] = v

        temp_zone_max = self._nonfluents['TEMP-ZONE-MAX']
        temp_zone_min = self._nonfluents['TEMP-ZONE-MIN']
        temp_out = self._nonfluents['TEMP-OUT']

        object_layout = {'temp-zone-max': temp_zone_max, 'temp-zone-min': temp_zone_min,
                         'temp-out': temp_out,
                         'zone-vol': zone_vol,
                         'heater-vol': heater_vol,
                         'temp-zone': temp_zone,
                         'temp-heater': temp_heater,
                         'adj-zone': adj_zone,
                         'adj-heater': adj_heater}

        return object_layout

    def init_canvas_info(self):
        interval = self._interval

        zone_list = self._objects['zone']
        heater_list = self._objects['heater']

        room_grid_size = (max(len(zone_list), len(heater_list)),
                          max(len(zone_list), len(heater_list)) + 1)

        room_center = (room_grid_size[0] * self._interval / 2,
                       (room_grid_size[1] - 1) * self._interval / 2 + interval)

        start_points_zone = []
        for i in range(room_grid_size[0]):
            for j in range(room_grid_size[1]):
                start_points_zone.append((i * interval, j * interval))

        start_points_heater = []
        for i in range(room_grid_size[0]):
            start_points_heater.append((i * interval, 0))

        # rank start point by distance from center
        start_points_zone.sort(
            key=lambda x: abs(x[0] + interval / 2 - room_center[0]) + \
                          abs(x[1] + interval / 2 - room_center[1]))
        start_points_heater.sort(
            key=lambda x: abs(x[0] + interval / 2 - room_center[0]) + \
                          abs(x[1] + interval / 2 - room_center[1]))

        # rank by number of adj rooms
        zone_list.sort(reverse=True,
                       key=lambda x: len(self._object_layout['adj-zone'][x]))
        heater_list.sort(reverse=True,
                         key=lambda x: len(self._object_layout['adj-heater'][x]))

        # print(zone_list, start_points_zone, room_center, room_grid_size)
        zone_init_points = {zone_list[i]: start_points_zone[i]
                            for i in range(len(zone_list))}
        heater_init_points = {heater_list[i]: start_points_heater[i]
                              for i in range(len(heater_list))}

        canvas_info = {
            'canvas_size': (room_grid_size[0] * interval,
                            room_grid_size[1] * interval),
            'zone_init_points': zone_init_points,
            'heater_init_points': heater_init_points
        }

        return canvas_info

    def render_conn(self):
        fig = self._fig
        ax = self._ax
        adj_zone = self._object_layout['adj-zone']
        adj_heater = self._object_layout['adj-heater']
        zone_init_points = self._canvas_info['zone_init_points']
        heater_init_points = self._canvas_info['heater_init_points']

        interval = self._interval * 2 / 3

        zone_line_list = []
        for k, v in adj_zone.items():
            from_point = (zone_init_points[k][0] + interval / 2,
                          zone_init_points[k][1] + interval / 2)
            for p in v:
                to_point = (zone_init_points[p][0] + interval / 2,
                            zone_init_points[p][1] + interval / 2)
                line = plt.Line2D((from_point[0], to_point[0]),
                                  (from_point[1], to_point[1]),
                                  ls='-', color='black', lw=2, zorder=1)
                ax.add_line(line)
        for k, v in adj_heater.items():
            from_point = (heater_init_points[k][0] + interval / 2,
                          heater_init_points[k][1] + interval / 2)
            for p in v:
                to_point = (zone_init_points[p][0] + interval / 2,
                            zone_init_points[p][1] + interval / 2)
                line = plt.Line2D((from_point[0], to_point[0]),
                                  (from_point[1], to_point[1]),
                                  ls='-', color='red', lw=2, zorder=1)
                ax.add_line(line)

    def render_zone(self, zone: str) -> tuple:

        fig = self._fig
        ax = self._ax

        curr_z = zone

        # vol_ratio = self._object_layout['zone-vol'][curr_z] / max(i for i in self._object_layout['zone-vol'].values())
        vol_ratio = 1

        init_x, init_y = self._canvas_info['zone_init_points'][curr_z]
        # center_x = init_x + self._interval / 2
        # center_y = init_y + self._interval / 2
        interval = self._interval * 2 / 3 * vol_ratio
        temp_ratio = 1

        temp = self._object_layout['temp-zone'][curr_z]
        upper_bound = self._object_layout['temp-zone-max']
        lower_bound = self._object_layout['temp-zone-min']

        upL = [init_x, upper_bound / 100 * interval + init_y]
        upR = [init_x + interval, upper_bound / 100 * interval + init_y]
        lowL = [init_x, lower_bound / 100 * interval + init_y]
        lowR = [init_x + interval, lower_bound / 100 * interval + init_y]

        background_rect = plt.Rectangle((init_x, init_y),
                                        interval,
                                        interval,
                                        fc='lightgrey', alpha=1, zorder=3)
        temp_rect = plt.Rectangle((init_x, init_y), interval, temp / 100 * interval, fc='crimson', zorder=3)
        line_up = plt.Line2D((upL[0], upR[0]), (upL[1], upR[1]), ls='--', color='orange', lw=0.5, zorder=3)
        line_low = plt.Line2D((lowL[0], lowR[0]), (lowL[1], lowR[1]), ls='--', color='lime', lw=0.5, zorder=4)

        ax.add_patch(background_rect)
        ax.add_patch(temp_rect)
        ax.add_line(line_up)
        ax.add_line(line_low)

        plt.text(init_x + interval * 0.8, init_y + interval * 0.8, "%s" % curr_z, color='black', fontsize=5, zorder=5)

        return

    def render_heater(self, heater: str) -> tuple:

        fig = self._fig
        ax = self._ax

        curr_h = heater

        # vol_ratio = self._object_layout['zone-vol'][curr_z] / max(i for i in self._object_layout['zone-vol'].values())
        vol_ratio = 1

        init_x, init_y = self._canvas_info['heater_init_points'][curr_h]
        interval = self._interval * 2 / 3 * vol_ratio
        temp_ratio = 1

        temp = self._object_layout['temp-heater'][curr_h]

        background_rect = plt.Rectangle((init_x, init_y),
                                        interval,
                                        interval,
                                        fc='lightcoral', alpha=1, zorder=3)
        temp_rect = plt.Rectangle((init_x, init_y), interval, temp / 100 * interval, fc='crimson', zorder=3)

        ax.add_patch(background_rect)
        ax.add_patch(temp_rect)

        plt.text(init_x + interval * 0.8, init_y + interval * 0.8, "%s" % curr_h, color='black', fontsize=5, zorder=5)

        return

    def fig2npa(self, fig):
        data = np.frombuffer(fig.canvas.tostring_rgb(), dtype=np.uint8)
        data = data.reshape(fig.canvas.get_width_height()[::-1] + (3,))
        return data

    def render(self, state) -> np.ndarray:

        # self._object_layout = self.build_object_layout()

        self._object_layout = self.build_object_layout()

        self._states = state

        self._canvas_info = self.init_canvas_info()

        canvas_size = self._canvas_info['canvas_size']

        self._fig = plt.figure(figsize=(canvas_size[0] / 10, canvas_size[1] / 10),
                               dpi=200)
        self._ax = plt.gca()

        plt.xlim([-canvas_size[0] / 5, canvas_size[0]])
        plt.ylim([-canvas_size[1] / 5, canvas_size[1]])
        plt.axis('scaled')
        plt.axis('off')

        for z in self._objects['zone']:
            self.render_zone(z)
        for h in self._objects['heater']:
            self.render_heater(h)

        self.render_conn()

        img = self.convert2img(self._fig, self._ax)

        self._ax.cla()
        plt.cla()
        plt.close()

        return img

        plt.show()
        sys.exit()

        # layout_data = []
        for res in self._objects['res']:
            curr_t = res
            fig, ax = self.render_res(curr_t)
            # layout_data.append(self.fig2npa(fig))
        self.render_conn()

        plt.show()
        sys.exit()

        # print(layout_data[0].shape)

        plt.imshow(layout_data[0])
        plt.axis('off')
        plt.show(block=False)
        plt.pause(20)
        plt.close()

    def convert2img(self, fig, ax):

        fig.canvas.draw()
        data = np.frombuffer(fig.canvas.tostring_rgb(), dtype=np.uint8)
        data = data.reshape(fig.canvas.get_width_height()[::-1] + (3,))
        img = Image.fromarray(data)
        self._data = data
        self._img = img

        return img



from pyRDDLGym.Core.Compiler.RDDLDecompiler import RDDLDecompiler
from pyRDDLGym.Core.Parser.expr import Expression

ERROR_MESSAGE_DECOMPILER = RDDLDecompiler()


def print_stack_trace(expr):
    if isinstance(expr, Expression):
        trace = ERROR_MESSAGE_DECOMPILER.decompile_expr(expr)
    else:
        trace = str(expr)
    return f'>> {trace}'


def print_stack_trace_root(expr, root):
    return print_stack_trace(expr) + '\n' + f'Please check expression for {root}.' 


class RDDLActionPreconditionNotSatisfiedError(ValueError):
    pass


class RDDLInvalidActionError(ValueError):
    pass


class RDDLInvalidDependencyInCPFError(SyntaxError):
    pass


class RDDLInvalidExpressionError(SyntaxError):
    pass


class RDDLInvalidNumberOfArgumentsError(SyntaxError):
    pass


class RDDLInvalidObjectError(SyntaxError):
    pass


class RDDLMissingCPFDefinitionError(SyntaxError):
    pass


class RDDLNotImplementedError(NotImplementedError):
    pass


class RDDLParseError(SyntaxError):
    pass


class RDDLRepeatedVariableError(SyntaxError):
    pass
    

class RDDLStateInvariantNotSatisfiedError(ValueError):
    pass


class RDDLTypeError(TypeError):
    pass


class RDDLUndefinedCPFError(SyntaxError):
    pass


class RDDLUndefinedVariableError(SyntaxError):
    pass


class RDDLValueOutOfRangeError(ValueError):
    pass


class RDDLEnvironmentNotExist(ValueError):
    pass


class RDDLInstanceNotExist(ValueError):
    pass


class RDDLLogFolderError(ValueError):
    pass


class RDDLEpisodeAlreadyEndedError(RuntimeError):
    pass



'''In this example, the PPO algorithm is used to solve a given domain.
    
The syntax for running this example is:

    python RLExample.py <domain> <instance> <steps>
    
where:
    <domain> is the name of a domain located in the /Examples directory
    <instance> is the instance number
    <steps> is the number of trials to simulate for training
'''
import sys
from stable_baselines3 import PPO

from pyRDDLGym.Core.Env.RDDLEnv import RDDLEnv
from pyRDDLGym.Examples.ExampleManager import ExampleManager

    
def main(domain, instance, steps):
    
    # set up the environment
    info = ExampleManager.GetEnvInfo(domain)    
    env = RDDLEnv.build(info, instance, enforce_action_constraints=True, 
                        new_gym_api=True, compact_action_space=True)
    env.reset()
    
    # train the agent
    model = PPO('MultiInputPolicy', env, verbose=1)
    model.learn(total_timesteps=steps)
    
    # evaluate
    total_reward = 0
    state, _ = env.reset()
    for step in range(env.horizon):
        env.render()
        action, _ = model.predict(state)
        next_state, reward, done, *_ = env.step(action)
        print(f'step       = {step}\n'
              f'state      = {state}\n'
              f'action     = {action}\n'
              f'next state = {next_state}\n'
              f'reward     = {reward}\n')
        total_reward += reward
        state = next_state
        if done:
            break
    print(f'episode ended with return {total_reward}')
    
    env.close()
        
        
if __name__ == "__main__":
    args = sys.argv[1:]
    if len(args) < 3:
        print('python RLExample.py <domain> <instance> <steps>')
        exit(1)
    domain, instance, steps = args[:3]
    main(domain, instance, steps)



import jax
import math
import numpy as np
import random
from typing import Dict, List

from pyRDDLGym.Core.Compiler.RDDLLiftedModel import RDDLLiftedModel
from pyRDDLGym.Core.Env.RDDLEnv import RDDLEnv
from pyRDDLGym.Core.Jax.JaxRDDLBackpropPlanner import JaxRDDLBackpropPlanner
from pyRDDLGym.Core.Jax.JaxRDDLBackpropPlanner import JaxStraightLinePlan
from pyRDDLGym.Core.Jax.JaxRDDLLogic import FuzzyLogic
from pyRDDLGym.Core.Jax.JaxRDDLSimulator import JaxRDDLSimulator

from pyRDDLGym.Examples.ExampleManager import ExampleManager


class ChanceNode:
    '''A chance node represents an action/decision in the search tree.'''
    
    def __init__(self, action: object) -> None:
        self.action = action
        self.children = []
    
    def add(self, vD: 'DecisionNode') -> None:
        self.children.append(vD)
        
    def lvn(self) -> 'DecisionNode':
        least, leastvisit = None, math.inf
        for c in self.children:
            score = c.N
            if score < leastvisit:
                least, leastvisit = c, score
        return least
                

class DecisionNode:
    '''A decision node represents a transition following some decision made.'''
    
    def __init__(self, transition: object, actions: List[object]=None) -> None:
        self.transition = transition
        self.N = 0
        self.Nc = []
        self.Qc = []
        self.children = []    
        if actions is not None:
            self.unvisited = actions.copy()
            random.shuffle(self.unvisited)
    
    def add(self, vC: 'ChanceNode') -> None:
        self.Nc.append(0)
        self.Qc.append(0.0)
        self.children.append(vC)
        
    def uct(self, c: float) -> int:
        logN = math.log(self.N)
        Qc, Nc = self.Qc, self.Nc
        best, bestscore = None, -math.inf
        for i in range(len(self.children)):
            score = Qc[i] + c * math.sqrt(logN / Nc[i])
            if score > bestscore:
                best, bestscore = i, score
        return best
    
    def update(self, i: int, R: float) -> None:
        self.N += 1
        self.Nc[i] += 1
        self.Qc[i] += (R - self.Qc[i]) / self.Nc[i]


class JaxRDDLHybridBackpropUCTPlanner:
    '''Represents a hybrid approach between UCT-MCTS and planning by backprop.
    '''
    
    def __init__(self, rddl: RDDLLiftedModel,
                 action_bounds: Dict,
                 rollout_horizon: int,
                 alpha: float,
                 beta: float,
                 delta: float,
                 c: float=1.0 / math.sqrt(2.0),
                 max_sgd_updates: int=1,
                 policy_hyperparams: Dict={},
                 **planner_kwargs) -> None:
        '''Creates a new hybrid backprop + UCT-MCTS planner.
        
        :param rddl: the rddl domain to optimize
        :param actions: the set of possible actions in the domain: can be None, 
        in which case progressive widening and the action_sample distribution are
        used instead
        :param rollout_horizon: how many steps to plan ahead
        :param alpha: the growth factor for chance nodes
        :param beta: the growth factor for decision nodes
        :param delta: how much can backprop updates change the MCTS plan
        :param c: scaling factor for UCT: note that this is adapted during learning
        so the parameter here specifies only the initial guess
        :param max_sgd_updates: max number of SGD updates for backprop planner
        :param policy_hyperparams: same as what would normally be passed to
        planner.optimize()
        :param **planner_kwargs: keywords arguments to initialize backprop planner:
        will not use backprop if none are specified 
        '''
        self.rddl = rddl
        self.sim = JaxRDDLSimulator(rddl, keep_tensors=True)  # TODO (mike): no need to compile twice
        self.T = rollout_horizon
        self.alpha = alpha
        self.beta = beta
        self.delta = delta
        self.c = c
        self.action_sample = self.action_bounds_to_samplers(rddl, action_bounds)     
        self.max_sgd_updates = max_sgd_updates
        
        # use progressive widening for action nodes if action space not specified
        self.use_pw_action = True
        
        # optionally use a backprop planner to refine the search tree
        if planner_kwargs is not None and planner_kwargs:
            self.sgd = JaxRDDLBackpropPlanner(
                rddl=rddl, 
                rollout_horizon=self.T, 
                action_bounds=action_bounds,
                **planner_kwargs)
        else:
            self.sgd = None
        self.policy_hyperparams = policy_hyperparams
    
    @staticmethod
    def action_bounds_to_samplers(rddl, action_bounds):
        samplers = {}
        for (action, prange) in rddl.actionsranges.items():
            ptypes = rddl.param_types[action]
            shape = rddl.object_counts(ptypes) if ptypes else None
            if prange == 'bool':
                samplers[action] = (lambda _: np.random.uniform(size=shape) < 0.5)
            else:
                bounds = action_bounds[action]
                samplers[action] = (lambda _: np.random.uniform(*bounds, size=shape))
        return samplers
        
    def _select_action(self, s, vD, t, c):
        
        # determine whether a new chance (action) node needs to be created
        # if so, proceed with roll-out stage after
        if self.use_pw_action:
            rollout = len(vD.children) <= vD.N ** self.alpha
        else:
            rollout = bool(vD.unvisited)
            
        # create a new chance node
        if rollout:
            
            # if progressive widening is used, sample action from the sampling
            # distribution, otherwise sample a random unvisited action
            if self.use_pw_action:
                a = {var: self.action_sample[var](s) for var in self.rddl.actions}
            else:
                a = vD.unvisited.pop()
                
            # append the new chance node to the decision node's children
            vC = ChanceNode(a)
            vD.add(vC)
            ivC = len(vD.children) - 1
        else:
            
            # sample an action from the UCT bound for the vD children
            ivC = vD.uct(c)
            vC = vD.children[ivC]
            a = vC.action     
            
        return (vC, ivC, a, rollout)
    
    def _sample_noise(self, vD, vC, ivC):
        
        # determine if a new decision (state) node needs to be created
        # for now assume continuous state space and use progressive widening
        if len(vC.children) <= vD.Nc[ivC] ** self.beta:
            
            # a new decision node must be created
            # here a new transition will be sampled from the environment
            vD1 = DecisionNode(None, None)
            vC.add(vD1)
        else:
            
            # just return the existing transition
            vD1 = vC.lvn()
        return vD1
    
    def _rollout(self, s, t):
        step_fn = self.sim.step
        policy = self.action_sample
        R = 0.0
        for _ in range(t, self.T):
            a = {var: policy[var](s) for var in self.rddl.actions}
            s, r, done = step_fn(a)
            R += r
            if done:
                break
        return R
    
    def _simulate(self, s, vD, t, c, vCs):
        
        # the end of the planning horizon
        if t >= self.T:
            return 0.0
        
        # select an action
        vC, ivC, a, rollout = self._select_action(s, vD, t, c)
        
        # perform a transition
        vD1 = self._sample_noise(vD, vC, ivC)
        tr = vD1.transition
        if tr is None:
            vD1.transition = tr = self.sim.step(a)
        s1, r, done = tr        
        vCs.append(vC)
        
        # termination early
        if done:
            t = self.T - 1
        
        # continue with rollout or simulation
        if rollout:
            R1 = self._rollout(s1, t + 1)
        else:
            R1 = self._simulate(s1, vD1, t + 1, c, vCs)
        
        # update decision node statistics
        R = r + R1
        vD.update(ivC, R)        
        return R
        
    def _jax_sgd(self, key, train_subs, test_subs, vCs): 
        
        # generate initial guess for plan from search tree
        sgd = self.sgd
        dtype = sgd.compiled.REAL
        action_names = vCs[0].action.keys()
        guess = {var: np.asarray([vC.action[var] for vC in vCs], dtype=dtype) 
                 for var in action_names}
        
        # run SGD updates
        model_params = sgd.compiled.model_params
        hyper_params = self.policy_hyperparams
        params = guess
        opt_state = sgd.optimizer.init(params)        
        for epoch in range(self.max_sgd_updates):
            key, subkey = jax.random.split(key)
            params, _, opt_state, _ = sgd.update(
                subkey, params, hyper_params, train_subs, model_params, opt_state)
            
            # stop once the actions diverge by delta
            max_delta = 0.0
            for var in action_names:
                delta = np.linalg.norm(params[var] - guess[var])
                max_delta = max(max_delta, delta)
            if max_delta >= self.delta:
                break
            
        # store update in the search tree
        num_steps = len(vCs)
        for step in range(num_steps):
            
            # get testing action
            action = sgd.test_policy(key, params, hyper_params, step, test_subs)
            
            # project action within a trust region around current estimate
            for var in action_names:
                if rddl.actionsranges[var] == 'real':
                    action[var] = np.clip(
                        action[var], guess[var] - self.delta, guess[var] + self.delta)
            vCs[step].action = action 
                   
        return max_delta, epoch + 1
        
    def search(self, key: jax.random.PRNGKey,
               s0: Dict,
               epochs: int,
               steps: int=1):
        '''Performs a search using the current planner.
        
        :param key: Jax PRNG key for random sampling
        :param s0: initial pvariables (e.g. state, non-fluents) to begin from
        :param epochs: how many iterations of MCTS to perform
        :param steps: how often to return a callback
        '''
        root = DecisionNode(None)
        R0 = self.c   
             
        sgd = self.sgd
        if sgd is not None:
            train_subs, test_subs = sgd._batched_init_subs(s0)
            
        for it in range(epochs):
            
            # regular MCTS with double-progressive widening
            self.sim.subs = s0.copy()
            vCs = []
            R = self._simulate(s0, root, 0, R0, vCs)
            R0 += (R - R0) / (it + 1)
            
            # backprop stage
            max_delta = np.nan
            if sgd is not None:
                key, subkey = jax.random.split(key)
                max_delta, num_updates = self._jax_sgd(
                    subkey, train_subs, test_subs, vCs)                
            
            # callback
            if it % steps == 0 or it == epochs - 1:
                best = root.uct(0.0)
                action = root.children[best].action
                yield {
                    'return': R,
                    'action': action,
                    'best': best,
                    'iter': it,
                    'c': R0,
                    'max_delta': max_delta,
                    'sgd_updates': num_updates,
                }


if __name__ == '__main__':
    EnvInfo = ExampleManager.GetEnvInfo('Wildfire')
    rddl = RDDLEnv(EnvInfo.get_domain(), EnvInfo.get_instance(0)).model
    
    world = JaxRDDLSimulator(rddl, keep_tensors=True)
    planner = JaxRDDLHybridBackpropUCTPlanner(
        rddl, 
        action_bounds={}, 
        rollout_horizon=5, 
        alpha=0.5, 
        beta=0.5, 
        delta=1.0,
        plan=JaxStraightLinePlan(),
        batch_size_train=32,
        policy_hyperparams={'cut-out': 5.0, 'put-out': 5.0},
        logic=FuzzyLogic(weight=100.),
        optimizer_kwargs={'learning_rate': 0.1})
    
    print('starting MCTS...')
    world.reset()
    key = jax.random.PRNGKey(42)
    total_reward = 0
    for step in range(100):        
        print('\n' + f'iteration {step}...')
        s0 = world.subs.copy()
        for callback in planner.search(key, s0, 200, 20):
            print(f'iter={callback["iter"]}, '
                  f'reward={callback["return"]}, '
                  f'c={callback["c"]}, '
                  f'delta={callback["max_delta"]}, '
                  f'updates={callback["sgd_updates"]}')
        _, reward, done = world.step(callback["action"])
        print(f'reward = {reward}')
        total_reward += reward
        if done:
            break
    print(f'total reward {total_reward}')
    


from ast import literal_eval
import configparser
import haiku as hk
import jax
import jax.numpy as jnp
import jax.random as random
import jax.nn.initializers as initializers
import numpy as np
np.seterr(all='raise')
import optax
import os
import sys
import time
from tqdm import tqdm
from typing import Callable, Dict, Generator, Set, Sequence, Tuple
import warnings

from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLNotImplementedError
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLTypeError

from pyRDDLGym.Core.Compiler.RDDLLiftedModel import RDDLLiftedModel
from pyRDDLGym.Core.Jax.JaxRDDLCompiler import JaxRDDLCompiler
from pyRDDLGym.Core.Jax import JaxRDDLLogic
from pyRDDLGym.Core.Jax.JaxRDDLLogic import FuzzyLogic
from pyRDDLGym.Core.Policies.Agents import BaseAgent


# ***********************************************************************
# CONFIG FILE MANAGEMENT
# 
# - read config files from file path
# - extract experiment settings
# - instantiate planner
#
# ***********************************************************************
def _parse_config_file(path: str):
    if not os.path.isfile(path):
        raise FileNotFoundError(f'File {path} does not exist.')
    config = configparser.RawConfigParser()
    config.optionxform = str 
    config.read(path)
    args = {k: literal_eval(v) 
            for section in config.sections()
            for (k, v) in config.items(section)}
    return config, args


def _parse_config_string(value: str):
    config = configparser.RawConfigParser()
    config.optionxform = str 
    config.read_string(value)
    args = {k: literal_eval(v) 
            for section in config.sections()
            for (k, v) in config.items(section)}
    return config, args


def _load_config(config, args):
    model_args = {k: args[k] for (k, _) in config.items('Model')}
    planner_args = {k: args[k] for (k, _) in config.items('Optimizer')}
    train_args = {k: args[k] for (k, _) in config.items('Training')}    
    
    train_args['key'] = jax.random.PRNGKey(train_args['key'])
    
    # read the model settings
    tnorm_name = model_args['tnorm']
    tnorm_kwargs = model_args['tnorm_kwargs']
    logic_name = model_args['logic']
    logic_kwargs = model_args['logic_kwargs']
    logic_kwargs['tnorm'] = getattr(JaxRDDLLogic, tnorm_name)(**tnorm_kwargs)
    planner_args['logic'] = getattr(JaxRDDLLogic, logic_name)(**logic_kwargs)
    
    # read the optimizer settings
    plan_method = planner_args.pop('method')
    plan_kwargs = planner_args.pop('method_kwargs', {})  
    
    if 'initializer' in plan_kwargs:  # weight initialization
        init_name = plan_kwargs['initializer']
        init_class = getattr(initializers, init_name)
        init_kwargs = plan_kwargs.pop('initializer_kwargs', {})
        try: 
            plan_kwargs['initializer'] = init_class(**init_kwargs)
        except:
            warnings.warn(f'ignoring arguments for initializer <{init_name}>',
                          stacklevel=2)
            plan_kwargs['initializer'] = init_class
               
    if 'activation' in plan_kwargs:  # activation function
        plan_kwargs['activation'] = getattr(jax.nn, plan_kwargs['activation'])
    
    planner_args['plan'] = getattr(sys.modules[__name__], plan_method)(**plan_kwargs)
    planner_args['optimizer'] = getattr(optax, planner_args['optimizer'])
    
    return planner_args, plan_kwargs, train_args


def load_config(path: str) -> Tuple[Dict[str, object], ...]:
    '''Loads a config file at the specified file path.'''
    config, args = _parse_config_file(path)
    return _load_config(config, args)


def load_config_from_string(value: str) -> Tuple[Dict[str, object], ...]:
    '''Loads config file contents specified explicitly as a string value.'''
    config, args = _parse_config_string(value)
    return _load_config(config, args)

    
# ***********************************************************************
# MODEL RELAXATIONS
# 
# - replace discrete ops in state dynamics/reward with differentiable ones
#
# ***********************************************************************
class JaxRDDLCompilerWithGrad(JaxRDDLCompiler):
    '''Compiles a RDDL AST representation to an equivalent JAX representation. 
    Unlike its parent class, this class treats all fluents as real-valued, and
    replaces all mathematical operations by equivalent ones with a well defined 
    (e.g. non-zero) gradient where appropriate. 
    '''
    
    def __init__(self, *args,
                 logic: FuzzyLogic=FuzzyLogic(),
                 cpfs_without_grad: Set=set(),
                 **kwargs) -> None:
        '''Creates a new RDDL to Jax compiler, where operations that are not
        differentiable are converted to approximate forms that have defined 
        gradients.
        
        :param *args: arguments to pass to base compiler
        :param logic: Fuzzy logic object that specifies how exact operations
        are converted to their approximate forms: this class may be subclassed
        to customize these operations
        :param cpfs_without_grad: which CPFs do not have gradients (use straight
        through gradient trick)
        :param *kwargs: keyword arguments to pass to base compiler
        '''
        super(JaxRDDLCompilerWithGrad, self).__init__(*args, **kwargs)
        self.logic = logic
        self.cpfs_without_grad = cpfs_without_grad
        
        # actions and CPFs must be continuous
        warnings.warn(f'Initial values of pvariables will be cast to real.',
                      stacklevel=2)   
        for (var, values) in self.init_values.items():
            self.init_values[var] = np.asarray(values, dtype=self.REAL) 
        
        # overwrite basic operations with fuzzy ones
        self.RELATIONAL_OPS = {
            '>=': logic.greaterEqual(),
            '<=': logic.lessEqual(),
            '<': logic.less(),
            '>': logic.greater(),
            '==': logic.equal(),
            '~=': logic.notEqual()
        }
        self.LOGICAL_NOT = logic.Not()
        self.LOGICAL_OPS = {
            '^': logic.And(),
            '&': logic.And(),
            '|': logic.Or(),
            '~': logic.xor(),
            '=>': logic.implies(),
            '<=>': logic.equiv()
        }
        self.AGGREGATION_OPS['forall'] = logic.forall()
        self.AGGREGATION_OPS['exists'] = logic.exists()
        self.AGGREGATION_OPS['argmin'] = logic.argmin()
        self.AGGREGATION_OPS['argmax'] = logic.argmax()
        self.KNOWN_UNARY['sgn'] = logic.signum()
        self.KNOWN_UNARY['floor'] = logic.floor()   
        self.KNOWN_UNARY['ceil'] = logic.ceil()   
        self.KNOWN_UNARY['round'] = logic.round()
        self.KNOWN_UNARY['sqrt'] = logic.sqrt()
        self.KNOWN_BINARY['div'] = logic.floorDiv()
        self.KNOWN_BINARY['mod'] = logic.mod()
        self.KNOWN_BINARY['fmod'] = logic.mod()
    
    def _jax_stop_grad(self, jax_expr):
        
        def _jax_wrapped_stop_grad(x, params, key):
            sample, key, error = jax_expr(x, params, key)
            sample = jax.lax.stop_gradient(sample)
            return sample, key, error
        
        return _jax_wrapped_stop_grad
        
    def _compile_cpfs(self, info):
        warnings.warn('CPFs outputs will be cast to real.', stacklevel=2)      
        jax_cpfs = {}
        for (_, cpfs) in self.levels.items():
            for cpf in cpfs:
                _, expr = self.rddl.cpfs[cpf]
                jax_cpfs[cpf] = self._jax(expr, info, dtype=self.REAL)
                if cpf in self.cpfs_without_grad:
                    warnings.warn(f'CPF <{cpf}> stops gradient.', stacklevel=2)      
                    jax_cpfs[cpf] = self._jax_stop_grad(jax_cpfs[cpf])
        return jax_cpfs
    
    def _jax_if_helper(self):
        return self.logic.If()
    
    def _jax_switch_helper(self):
        return self.logic.Switch()
        
    def _jax_kron(self, expr, info):
        if self.logic.verbose:
            warnings.warn('KronDelta will be ignored.', stacklevel=2)            
                       
        arg, = expr.args
        arg = self._jax(arg, info)
        return arg
    
    def _jax_bernoulli_helper(self):
        return self.logic.bernoulli()
    
    def _jax_discrete_helper(self):
        jax_discrete, jax_param = self.logic.discrete()

        def _jax_wrapped_discrete_calc_approx(key, prob, params):
            sample = jax_discrete(key, prob, params)
            out_of_bounds = jnp.logical_not(jnp.logical_and(
                jnp.all(prob >= 0),
                jnp.allclose(jnp.sum(prob, axis=-1), 1.0)))
            return sample, out_of_bounds
        
        return _jax_wrapped_discrete_calc_approx, jax_param


# ***********************************************************************
# ALL VERSIONS OF JAX PLANS
# 
# - straight line plan
# - deep reactive policy
#
# ***********************************************************************
class JaxPlan:
    '''Base class for all JAX policy representations.'''
    
    def __init__(self) -> None:
        self._initializer = None
        self._train_policy = None
        self._test_policy = None
        self._projection = None
    
    def summarize_hyperparameters(self):
        pass
        
    def compile(self, compiled: JaxRDDLCompilerWithGrad,
                _bounds: Dict,
                horizon: int) -> None:
        raise NotImplementedError
    
    def guess_next_epoch(self, params: Dict) -> Dict:
        raise NotImplementedError
    
    @property
    def initializer(self):
        return self._initializer

    @initializer.setter
    def initializer(self, value):
        self._initializer = value
    
    @property
    def train_policy(self):
        return self._train_policy

    @train_policy.setter
    def train_policy(self, value):
        self._train_policy = value
        
    @property
    def test_policy(self):
        return self._test_policy

    @test_policy.setter
    def test_policy(self, value):
        self._test_policy = value
         
    @property
    def projection(self):
        return self._projection

    @projection.setter
    def projection(self, value):
        self._projection = value
    
    def _calculate_action_info(self, compiled: JaxRDDLCompilerWithGrad,
                               user_bounds: Dict[str, object], horizon: int):
        shapes, bounds, bounds_safe, cond_lists = {}, {}, {}, {}
        for (name, prange) in compiled.rddl.variable_ranges.items():
            if compiled.rddl.variable_types[name] != 'action-fluent':
                continue
            
            if prange not in compiled.JAX_TYPES:
                raise RDDLTypeError(
                    f'Invalid range <{prange}. of action-fluent <{name}>, '
                    f'must be one of {set(compiled.JAX_TYPES.keys())}.')
                
            # clip boolean to (0, 1), otherwise use the RDDL action bounds
            # or the user defined action bounds if provided
            shapes[name] = (horizon,) + np.shape(compiled.init_values[name])
            if prange == 'bool':
                lower, upper = None, None
            else:
                lower, upper = compiled.constraints.bounds[name]
                lower, upper = user_bounds.get(name, (lower, upper))
                lower = np.asarray(lower, dtype=np.float32)
                upper = np.asarray(upper, dtype=np.float32)
                lower_finite = np.isfinite(lower)
                upper_finite = np.isfinite(upper)
                bounds_safe[name] = (np.where(lower_finite, lower, 0.0),
                                     np.where(upper_finite, upper, 0.0))
                cond_lists[name] = [lower_finite & upper_finite,
                                    lower_finite & ~upper_finite,
                                    ~lower_finite & upper_finite,
                                    ~lower_finite & ~upper_finite]
            bounds[name] = (lower, upper)
            warnings.warn(f'Bounds of action fluent <{name}> set to '
                          f'{bounds[name]}', stacklevel=2)
        return shapes, bounds, bounds_safe, cond_lists
    
    def _count_bool_actions(self, rddl: RDDLLiftedModel):
        constraint = rddl.max_allowed_actions
        num_bool_actions = sum(np.size(values)
                               for (var, values) in rddl.actions.items()
                               if rddl.variable_ranges[var] == 'bool')
        return num_bool_actions, constraint

    
class JaxStraightLinePlan(JaxPlan):
    '''A straight line plan implementation in JAX'''
    
    def __init__(self, initializer: initializers.Initializer=initializers.normal(),
                 wrap_sigmoid: bool=True,
                 min_action_prob: float=1e-5,
                 wrap_non_bool: bool=False,
                 wrap_softmax: bool=False,
                 use_new_projection: bool=False,
                 max_constraint_iter: int=100) -> None:
        '''Creates a new straight line plan in JAX.
        
        :param initializer: a Jax Initializer for setting the initial actions
        :param wrap_sigmoid: wrap bool action parameters with sigmoid 
        (uses gradient clipping instead of sigmoid if None; this flag is ignored
        if wrap_softmax = True)
        :param min_action_prob: minimum value a soft boolean action can take
        (maximum is 1 - min_action_prob); required positive if wrap_sigmoid = True
        :param wrap_non_bool: whether to wrap real or int action fluent parameters
        with non-linearity (e.g. sigmoid or ELU) to satisfy box constraints
        :param wrap_softmax: whether to use softmax activation approach 
        (note, this is limited to max-nondef-actions = 1) instead of projected
        gradient to satisfy action constraints 
        :param use_new_projection: whether to use non-iterative (e.g. sort-based)
        projection method, or modified SOGBOFA projection method to satisfy
        action concurrency constraint
        :param max_constraint_iter: max iterations of projected 
        gradient for ensuring actions satisfy constraints, only required if 
        use_new_projection = True
        '''
        super(JaxStraightLinePlan, self).__init__()
        self._initializer_base = initializer
        self._initializer = initializer
        self._wrap_sigmoid = wrap_sigmoid
        self._min_action_prob = min_action_prob
        self._wrap_non_bool = wrap_non_bool
        self._wrap_softmax = wrap_softmax
        self._use_new_projection = use_new_projection
        self._max_constraint_iter = max_constraint_iter
        
    def summarize_hyperparameters(self):
        print(f'policy hyper-parameters:\n'
              f'    initializer          ={type(self._initializer_base).__name__}\n'
              f'constraint-sat strategy (simple):\n'
              f'    wrap_sigmoid         ={self._wrap_sigmoid}\n'
              f'    wrap_sigmoid_min_prob={self._min_action_prob}\n'
              f'    wrap_non_bool        ={self._wrap_non_bool}\n'
              f'constraint-sat strategy (complex):\n'
              f'    wrap_softmax         ={self._wrap_softmax}\n'
              f'    use_new_projection   ={self._use_new_projection}')
    
    def compile(self, compiled: JaxRDDLCompilerWithGrad,
                _bounds: Dict, horizon: int) -> None:
        rddl = compiled.rddl
        
        # calculate the correct action box bounds
        shapes, bounds, bounds_safe, cond_lists = self._calculate_action_info(
            compiled, _bounds, horizon)
        self.bounds = bounds
        
        # action concurrency check
        bool_action_count, allowed_actions = self._count_bool_actions(rddl)
        use_constraint_satisfaction = allowed_actions < bool_action_count        
        if use_constraint_satisfaction: 
            warnings.warn(f'Using projected gradient trick to satisfy '
                          f'max_nondef_actions: total boolean actions '
                          f'{bool_action_count} > max_nondef_actions '
                          f'{allowed_actions}.', stacklevel=2)
            
        noop = {var: (values[0] if isinstance(values, list) else values)
                for (var, values) in rddl.actions.items()}
        bool_key = 'bool__'
        
        # ***********************************************************************
        # STRAIGHT-LINE PLAN
        #
        # ***********************************************************************
        
        # define the mapping between trainable parameter and action
        wrap_sigmoid = self._wrap_sigmoid
        bool_threshold = 0.0 if wrap_sigmoid else 0.5
        
        def _jax_bool_param_to_action(var, param, hyperparams):
            if wrap_sigmoid:
                weight = hyperparams[var]
                return jax.nn.sigmoid(weight * param)
            else:
                return param 
        
        def _jax_bool_action_to_param(var, action, hyperparams):
            if wrap_sigmoid:
                weight = hyperparams[var]
                return (-1.0 / weight) * jnp.log1p(1.0 / action - 2.0)
            else:
                return action
            
        wrap_non_bool = self._wrap_non_bool
        
        def _jax_non_bool_param_to_action(var, param, hyperparams):
            if wrap_non_bool:
                lower, upper = bounds_safe[var]
                action = jnp.select(
                    condlist=cond_lists[var],
                    choicelist=[
                        lower + (upper - lower) * jax.nn.sigmoid(param),
                        lower + (jax.nn.elu(param) + 1.0),
                        upper - (jax.nn.elu(-param) + 1.0),
                        param
                    ]
                )
            else:
                action = param
            return action
        
        # handle box constraints    
        min_action = self._min_action_prob
        max_action = 1.0 - min_action
        
        def _jax_project_bool_to_box(var, param, hyperparams):
            lower = _jax_bool_action_to_param(var, min_action, hyperparams)
            upper = _jax_bool_action_to_param(var, max_action, hyperparams)
            valid_param = jnp.clip(param, lower, upper)
            return valid_param
        
        ranges = rddl.variable_ranges
        
        def _jax_wrapped_slp_project_to_box(params, hyperparams):
            new_params = {}
            for (var, param) in params.items():
                if var == bool_key:
                    new_params[var] = param
                elif ranges[var] == 'bool':
                    new_params[var] = _jax_project_bool_to_box(var, param, hyperparams)
                elif wrap_non_bool:
                    new_params[var] = param
                else:
                    new_params[var] = jnp.clip(param, *bounds[var])
            return new_params, True
        
        # convert softmax action back to action dict
        action_sizes = {var: np.prod(shape[1:], dtype=int) 
                        for (var, shape) in shapes.items()
                        if ranges[var] == 'bool'}
        
        def _jax_unstack_bool_from_softmax(output):
            actions = {}
            start = 0
            for (name, size) in action_sizes.items():
                action = output[..., start:start + size]
                action = jnp.reshape(action, newshape=shapes[name][1:])
                if noop[name]:
                    action = 1.0 - action
                actions[name] = action
                start += size
            return actions
                
        # train plan prediction (TODO: implement one-hot for integer actions)        
        def _jax_wrapped_slp_predict_train(key, params, hyperparams, step, subs):
            actions = {}
            for (var, param) in params.items():
                action = jnp.asarray(param[step, ...], dtype=compiled.REAL)
                if var == bool_key:
                    output = jax.nn.softmax(action)
                    bool_actions = _jax_unstack_bool_from_softmax(output)
                    actions.update(bool_actions)
                elif ranges[var] == 'bool':
                    actions[var] = _jax_bool_param_to_action(var, action, hyperparams)
                else:
                    actions[var] = _jax_non_bool_param_to_action(var, action, hyperparams)
            return actions
        
        # test plan prediction
        def _jax_wrapped_slp_predict_test(key, params, hyperparams, step, subs):
            actions = {}
            for (var, param) in params.items():
                action = jnp.asarray(param[step, ...])
                if var == bool_key:
                    output = jax.nn.softmax(action)
                    bool_actions = _jax_unstack_bool_from_softmax(output)
                    for (bool_var, bool_action) in bool_actions.items():
                        actions[bool_var] = bool_action > 0.5
                elif ranges[var] == 'bool':
                    actions[var] = action > bool_threshold
                else:
                    action = _jax_non_bool_param_to_action(var, action, hyperparams)
                    action = jnp.clip(action, *bounds[var])
                    if ranges[var] == 'int':
                        action = jnp.round(action).astype(compiled.INT)
                    actions[var] = action
            return actions
        
        self.train_policy = _jax_wrapped_slp_predict_train
        self.test_policy = _jax_wrapped_slp_predict_test
        
        # ***********************************************************************
        # ACTION CONSTRAINT SATISFACTION
        #
        # ***********************************************************************
        
        # use a softmax output activation
        if use_constraint_satisfaction and self._wrap_softmax:
            
            # only allow one action non-noop for now
            if 1 < allowed_actions < bool_action_count:
                raise RDDLNotImplementedError(
                    f'Straight-line plans with wrap_softmax currently '
                    f'do not support max-nondef-actions = {allowed_actions} > 1.')
                
            # potentially apply projection but to non-bool actions only
            self.projection = _jax_wrapped_slp_project_to_box
            
        # use new gradient projection method...
        elif use_constraint_satisfaction and self._use_new_projection:
            
            # shift the boolean actions uniformly, clipping at the min/max values
            # the amount to move is such that only top allowed_actions actions
            # are still active (e.g. not equal to noop) after the shift
            def _jax_wrapped_sorting_project(params, hyperparams):
                
                # find the amount to shift action parameters
                # if noop is True pretend it is False and reflect the parameter
                scores = []
                for (var, param) in params.items():
                    if ranges[var] == 'bool':
                        param_flat = jnp.ravel(param)
                        if noop[var]:
                            param_flat = (-param_flat) if wrap_sigmoid else 1.0 - param_flat
                        scores.append(param_flat)
                scores = jnp.concatenate(scores)
                descending = jnp.sort(scores)[::-1]
                kplus1st_greatest = descending[allowed_actions]
                surplus = jnp.maximum(kplus1st_greatest - bool_threshold, 0.0)
                    
                # perform the shift
                new_params = {}
                for (var, param) in params.items():
                    if ranges[var] == 'bool':
                        new_param = param + (surplus if noop[var] else -surplus)
                        new_param = _jax_project_bool_to_box(var, new_param, hyperparams)
                    else:
                        new_param = param
                    new_params[var] = new_param
                return new_params, True
                
            # clip actions to valid bounds and satisfy constraint on max actions
            def _jax_wrapped_slp_project_to_max_constraint(params, hyperparams):
                params, _ = _jax_wrapped_slp_project_to_box(params, hyperparams)
                project_over_horizon = jax.vmap(
                    _jax_wrapped_sorting_project, in_axes=(0, None)
                )(params, hyperparams)
                return project_over_horizon
            
            self.projection = _jax_wrapped_slp_project_to_max_constraint
        
        # use SOGBOFA projection method...
        elif use_constraint_satisfaction and not self._use_new_projection:
            
            # calculate the surplus of actions above max-nondef-actions
            def _jax_wrapped_sogbofa_surplus(params, hyperparams):
                sum_action, count = 0.0, 0
                for (var, param) in params.items():
                    if ranges[var] == 'bool':
                        action = _jax_bool_param_to_action(var, param, hyperparams)                        
                        if noop[var]:
                            sum_action += jnp.size(action) - jnp.sum(action)
                            count += jnp.sum(action < 1)
                        else:
                            sum_action += jnp.sum(action)
                            count += jnp.sum(action > 0)
                surplus = jnp.maximum(sum_action - allowed_actions, 0.0)
                count = jnp.maximum(count, 1)
                return surplus / count
                
            # return whether the surplus is positive or reached compute limit
            max_constraint_iter = self._max_constraint_iter
        
            def _jax_wrapped_sogbofa_continue(values):
                it, _, _, surplus = values
                return jnp.logical_and(it < max_constraint_iter, surplus > 0)
                
            # reduce all bool action values by the surplus clipping at minimum
            # for no-op = True, do the opposite, i.e. increase all
            # bool action values by surplus clipping at maximum
            def _jax_wrapped_sogbofa_subtract_surplus(values):
                it, params, hyperparams, surplus = values
                new_params = {}
                for (var, param) in params.items():
                    if ranges[var] == 'bool':
                        action = _jax_bool_param_to_action(var, param, hyperparams)
                        new_action = action + (surplus if noop[var] else -surplus)
                        new_action = jnp.clip(new_action, min_action, max_action)
                        new_param = _jax_bool_action_to_param(var, new_action, hyperparams)
                    else:
                        new_param = param
                    new_params[var] = new_param
                new_surplus = _jax_wrapped_sogbofa_surplus(new_params, hyperparams)
                new_it = it + 1
                return new_it, new_params, hyperparams, new_surplus
                
            # apply the surplus to the actions until it becomes zero
            def _jax_wrapped_sogbofa_project(params, hyperparams):
                surplus = _jax_wrapped_sogbofa_surplus(params, hyperparams)
                _, params, _, surplus = jax.lax.while_loop(
                    cond_fun=_jax_wrapped_sogbofa_continue,
                    body_fun=_jax_wrapped_sogbofa_subtract_surplus,
                    init_val=(0, params, hyperparams, surplus)
                )
                converged = jnp.logical_not(surplus > 0)
                return params, converged
                
            # clip actions to valid bounds and satisfy constraint on max actions
            def _jax_wrapped_slp_project_to_max_constraint(params, hyperparams):
                params, _ = _jax_wrapped_slp_project_to_box(params, hyperparams)
                project_over_horizon = jax.vmap(
                    _jax_wrapped_sogbofa_project, in_axes=(0, None)
                )(params, hyperparams)
                return project_over_horizon
            
            self.projection = _jax_wrapped_slp_project_to_max_constraint
        
        # just project to box constraints
        else: 
            self.projection = _jax_wrapped_slp_project_to_box
            
        # ***********************************************************************
        # PLAN INITIALIZATION
        #
        # ***********************************************************************
        
        init = self._initializer
        stack_bool_params = use_constraint_satisfaction and self._wrap_softmax
        
        def _jax_wrapped_slp_init(key, hyperparams, subs):
            params = {}
            for (var, shape) in shapes.items():
                if ranges[var] != 'bool' or not stack_bool_params: 
                    key, subkey = random.split(key)
                    param = init(subkey, shape, dtype=compiled.REAL)
                    if ranges[var] == 'bool':
                        param += bool_threshold
                    params[var] = param
            if stack_bool_params:
                key, subkey = random.split(key)
                bool_shape = (horizon, bool_action_count)
                bool_param = init(subkey, bool_shape, dtype=compiled.REAL)
                params[bool_key] = bool_param
            params, _ = _jax_wrapped_slp_project_to_box(params, hyperparams)
            return params
        
        self.initializer = _jax_wrapped_slp_init
    
    @staticmethod
    @jax.jit
    def _guess_next_epoch(param):
        # "progress" the plan one step forward and set last action to second-last
        return jnp.append(param[1:, ...], param[-1:, ...], axis=0)

    def guess_next_epoch(self, params: Dict) -> Dict:
        next_fn = JaxStraightLinePlan._guess_next_epoch
        return jax.tree_map(next_fn, params)


class JaxDeepReactivePolicy(JaxPlan):
    '''A deep reactive policy network implementation in JAX.'''
    
    def __init__(self, topology: Sequence[int],
                 activation: Callable=jax.nn.relu,
                 initializer: hk.initializers.Initializer=hk.initializers.VarianceScaling(scale=2.0),
                 normalize: bool=True) -> None:
        '''Creates a new deep reactive policy in JAX.
        
        :param neurons: sequence consisting of the number of neurons in each
        layer of the policy
        :param activation: function to apply after each layer of the policy
        :param initializer: weight initialization
        :param normalize: whether to apply layer norm to the inputs
        '''
        super(JaxDeepReactivePolicy, self).__init__()
        self._topology = topology
        self._activations = [activation for _ in topology]
        self._initializer_base = initializer
        self._initializer = initializer
        self._normalize = normalize
            
    def summarize_hyperparameters(self):
        print(f'policy hyper-parameters:\n'
              f'    topology        ={self._topology}\n'
              f'    activation_fn   ={self._activations[0].__name__}\n'
              f'    initializer     ={type(self._initializer_base).__name__}\n'
              f'    apply_layer_norm={self._normalize}')
    
    def compile(self, compiled: JaxRDDLCompilerWithGrad,
                _bounds: Dict, horizon: int) -> None:
        rddl = compiled.rddl
        
        # calculate the correct action box bounds
        shapes, bounds, bounds_safe, cond_lists = self._calculate_action_info(
            compiled, _bounds, horizon)
        shapes = {var: value[1:] for (var, value) in shapes.items()}
        self.bounds = bounds
        
        # action concurrency check - only allow one action non-noop for now
        bool_action_count, allowed_actions = self._count_bool_actions(rddl)
        if 1 < allowed_actions < bool_action_count:
            raise RDDLNotImplementedError(
                f'Deep reactive policies currently do not support '
                f'max-nondef-actions = {allowed_actions} > 1.')
        use_constraint_satisfaction = allowed_actions < bool_action_count
            
        noop = {var: (values[0] if isinstance(values, list) else values)
                for (var, values) in rddl.actions.items()}                   
        bool_key = 'bool__'
        
        # ***********************************************************************
        # POLICY NETWORK PREDICTION
        #
        # ***********************************************************************
                   
        ranges = rddl.variable_ranges
        normalize = self._normalize
        init = self._initializer
        layers = list(enumerate(zip(self._topology, self._activations)))
        layer_sizes = {var: np.prod(shape, dtype=int) 
                       for (var, shape) in shapes.items()}
        layer_names = {var: f'output_{var}'.replace('-', '_') for var in shapes}
        
        # predict actions from the policy network for current state
        def _jax_wrapped_policy_network_predict(state):
            
            # apply layer norm
            if normalize:
                normalizer = hk.LayerNorm(
                    axis=-1, param_axis=-1,
                    create_offset=True, create_scale=True,
                    name='input_norm')
                state = normalizer(state)
            
            # feed state vector through hidden layers
            hidden = state
            for (i, (num_neuron, activation)) in layers:
                linear = hk.Linear(num_neuron, name=f'hidden_{i}', w_init=init)
                hidden = activation(linear(hidden))
            
            # each output is a linear layer reshaped to original lifted shape
            actions = {}
            for (var, size) in layer_sizes.items():
                linear = hk.Linear(size, name=layer_names[var], w_init=init)
                reshape = hk.Reshape(output_shape=shapes[var], preserve_dims=-1,
                                     name=f'reshape_{layer_names[var]}')
                output = reshape(linear(hidden))
                if not shapes[var]:
                    output = jnp.squeeze(output)
                
                # project action output to valid box constraints 
                if ranges[var] == 'bool':
                    if not use_constraint_satisfaction:
                        actions[var] = jax.nn.sigmoid(output)
                else:
                    lower, upper = bounds_safe[var]
                    action = jnp.select(
                        condlist=cond_lists[var],
                        choicelist=[
                            lower + (upper - lower) * jax.nn.sigmoid(output),
                            lower + (jax.nn.elu(output) + 1.0),
                            upper - (jax.nn.elu(-output) + 1.0),
                            output
                        ]
                    )
                    actions[var] = action
            
            # for constraint satisfaction wrap bool actions with softmax
            if use_constraint_satisfaction:
                linear = hk.Linear(
                    bool_action_count, name='output_bool', w_init=init)
                output = jax.nn.softmax(linear(hidden))
                actions[bool_key] = output
             
            return actions
        
        predict_fn = hk.transform(_jax_wrapped_policy_network_predict)
        predict_fn = hk.without_apply_rng(predict_fn)            
        
        # convert softmax action back to action dict
        def _jax_unstack_bool_from_softmax(output):
            actions = {}
            start = 0
            for (name, size) in layer_sizes.items():
                if ranges[name] == 'bool':
                    action = output[..., start:start + size]
                    action = jnp.reshape(action, newshape=shapes[name])
                    if noop[name]:
                        action = 1.0 - action
                    actions[name] = action
                    start += size
            return actions
                
        # state is concatenated into single tensor
        def _jax_wrapped_subs_to_state(subs):
            subs = {var: value
                    for (var, value) in subs.items()
                    if var in rddl.states}
            flat_subs = jax.tree_map(jnp.ravel, subs)
            flat_subs = jax.tree_map(jnp.atleast_1d, flat_subs)
            states = list(flat_subs.values())
            state = jnp.concatenate(states)
            return state
        
        # train action prediction
        def _jax_wrapped_drp_predict_train(key, params, hyperparams, step, subs):
            state = _jax_wrapped_subs_to_state(subs)
            actions = predict_fn.apply(params, state)
            if use_constraint_satisfaction:
                bool_actions = _jax_unstack_bool_from_softmax(actions[bool_key])
                actions.update(bool_actions)
                del actions[bool_key]
            return actions
        
        # test action prediction
        def _jax_wrapped_drp_predict_test(key, params, hyperparams, step, subs):
            actions = _jax_wrapped_drp_predict_train(
                key, params, hyperparams, step, subs)
            new_actions = {}
            for (var, action) in actions.items():
                prange = ranges[var]
                if prange == 'bool':
                    new_action = action > 0.5
                elif prange == 'int':
                    action = jnp.clip(action, *bounds[var])
                    new_action = jnp.round(action).astype(compiled.INT)
                else:
                    new_action = jnp.clip(action, *bounds[var])
                new_actions[var] = new_action
            return new_actions
        
        self.train_policy = _jax_wrapped_drp_predict_train
        self.test_policy = _jax_wrapped_drp_predict_test
        
        # ***********************************************************************
        # ACTION CONSTRAINT SATISFACTION
        #
        # ***********************************************************************
        
        # no projection applied since the actions are already constrained
        def _jax_wrapped_drp_no_projection(params, hyperparams):
            return params, True
        
        self.projection = _jax_wrapped_drp_no_projection
    
        # ***********************************************************************
        # POLICY NETWORK INITIALIZATION
        #
        # ***********************************************************************
        
        def _jax_wrapped_drp_init(key, hyperparams, subs):
            subs = {var: value[0, ...] 
                    for (var, value) in subs.items()
                    if var in rddl.states}
            state = _jax_wrapped_subs_to_state(subs)
            params = predict_fn.init(key, state)
            return params
        
        self.initializer = _jax_wrapped_drp_init
        
    def guess_next_epoch(self, params: Dict) -> Dict:
        return params

    
# ***********************************************************************
# ALL VERSIONS OF JAX PLANNER
# 
# - simple gradient descent based planner
# - more stable but slower line search based planner
#
# ***********************************************************************
class JaxRDDLBackpropPlanner:
    '''A class for optimizing an action sequence in the given RDDL MDP using 
    gradient descent.'''
    
    def __init__(self, rddl: RDDLLiftedModel,
                 plan: JaxPlan,
                 batch_size_train: int,
                 batch_size_test: int=None,
                 rollout_horizon: int=None,
                 use64bit: bool=False,
                 action_bounds: Dict[str, Tuple[np.ndarray, np.ndarray]]={},
                 optimizer: Callable[..., optax.GradientTransformation]=optax.rmsprop,
                 optimizer_kwargs: Dict[str, object]={'learning_rate': 0.1},
                 clip_grad: float=None,
                 logic: FuzzyLogic=FuzzyLogic(),
                 use_symlog_reward: bool=False,
                 utility=jnp.mean,
                 cpfs_without_grad: Set=set()) -> None:
        '''Creates a new gradient-based algorithm for optimizing action sequences
        (plan) in the given RDDL. Some operations will be converted to their
        differentiable counterparts; the specific operations can be customized
        by providing a subclass of FuzzyLogic.
        
        :param rddl: the RDDL domain to optimize
        :param plan: the policy/plan representation to optimize
        :param batch_size_train: how many rollouts to perform per optimization 
        step
        :param batch_size_test: how many rollouts to use to test the plan at each
        optimization step
        :param rollout_horizon: lookahead planning horizon: None uses the
        :param use64bit: whether to perform arithmetic in 64 bit
        horizon parameter in the RDDL instance
        :param action_bounds: box constraints on actions
        :param optimizer: a factory for an optax SGD algorithm
        :param optimizer_kwargs: a dictionary of parameters to pass to the SGD
        factory (e.g. which parameters are controllable externally)
        :param clip_grad: maximum magnitude of gradient updates
        :param logic: a subclass of FuzzyLogic for mapping exact mathematical
        operations to their differentiable counterparts 
        :param use_symlog_reward: whether to use the symlog transform on the 
        reward as a form of normalization
        :param utility: how to aggregate return observations to compute utility
        of a policy or plan
        :param cpfs_without_grad: which CPFs do not have gradients (use straight
        through gradient trick)
        '''
        self.rddl = rddl
        self.plan = plan
        self.batch_size_train = batch_size_train
        if batch_size_test is None:
            batch_size_test = batch_size_train
        self.batch_size_test = batch_size_test
        if rollout_horizon is None:
            rollout_horizon = rddl.horizon
        self.horizon = rollout_horizon
        self._action_bounds = action_bounds
        self.use64bit = use64bit
        self._optimizer_name = optimizer
        self._optimizer_kwargs = optimizer_kwargs
        self.clip_grad = clip_grad
        
        # set optimizer
        try:
            optimizer = optax.inject_hyperparams(optimizer)(**optimizer_kwargs)
        except:
            warnings.warn(
                'Failed to inject hyperparameters into optax optimizer, '
                'rolling back to safer method: please note that modification of '
                'optimizer hyperparameters will not work, and it is '
                'recommended to update your packages and Python distribution.',
                stacklevel=2)
            optimizer = optimizer(**optimizer_kwargs)     
        if clip_grad is None:
            self.optimizer = optimizer
        else:
            self.optimizer = optax.chain(
                optax.clip(clip_grad),
                optimizer
            )
            
        self.logic = logic
        self.use_symlog_reward = use_symlog_reward
        self.utility = utility
        self.cpfs_without_grad = cpfs_without_grad
        
        self._jax_compile_rddl()        
        self._jax_compile_optimizer()
        
    def summarize_hyperparameters(self):
        print(f'objective and relaxations:\n'
              f'    objective_fn    ={self.utility.__name__}\n'
              f'    use_symlog      ={self.use_symlog_reward}\n'
              f'    lookahead       ={self.horizon}\n'
              f'    model relaxation={type(self.logic).__name__}\n'
              f'    action_bounds   ={self._action_bounds}\n'
              f'    cpfs_no_gradient={self.cpfs_without_grad}\n'
              f'optimizer hyper-parameters:\n'
              f'    use_64_bit      ={self.use64bit}\n'
              f'    optimizer       ={self._optimizer_name.__name__}\n'
              f'    optimizer args  ={self._optimizer_kwargs}\n'
              f'    clip_gradient   ={self.clip_grad}\n'
              f'    batch_size_train={self.batch_size_train}\n'
              f'    batch_size_test ={self.batch_size_test}')
        self.plan.summarize_hyperparameters()
        self.logic.summarize_hyperparameters()
        
    def _jax_compile_rddl(self):
        rddl = self.rddl
        
        # Jax compilation of the differentiable RDDL for training
        self.compiled = JaxRDDLCompilerWithGrad(
            rddl=rddl,
            logic=self.logic,
            use64bit=self.use64bit,
            cpfs_without_grad=self.cpfs_without_grad)
        self.compiled.compile()
        
        # Jax compilation of the exact RDDL for testing
        self.test_compiled = JaxRDDLCompiler(
            rddl=rddl,
            use64bit=self.use64bit)
        self.test_compiled.compile()
    
        # calculate grounded no-op actions
        self.noop_actions = {}
        for (var, values) in self.test_compiled.init_values.items():
            if rddl.variable_types[var] == 'action-fluent':
                self.noop_actions.update(rddl.ground_values(var, values))
        
    def _jax_compile_optimizer(self):
        
        # policy
        self.plan.compile(self.compiled,
                          _bounds=self._action_bounds,
                          horizon=self.horizon)
        self.train_policy = jax.jit(self.plan.train_policy)
        self.test_policy = jax.jit(self.plan.test_policy)
        
        # roll-outs
        train_rollouts = self.compiled.compile_rollouts(
            policy=self.plan.train_policy,
            n_steps=self.horizon,
            n_batch=self.batch_size_train)
        
        test_rollouts = self.test_compiled.compile_rollouts(
            policy=self.plan.test_policy,
            n_steps=self.horizon,
            n_batch=self.batch_size_test)
        self.test_rollouts = jax.jit(test_rollouts)
        
        # initialization
        self.initialize = jax.jit(self._jax_init())
        
        # losses
        train_loss = self._jax_loss(train_rollouts, use_symlog=self.use_symlog_reward)
        self.train_loss = jax.jit(train_loss)
        self.test_loss = jax.jit(self._jax_loss(test_rollouts, use_symlog=False))
        
        # optimization
        self.update = jax.jit(self._jax_update(train_loss))
    
    def _jax_wrapped_returns_fn(self, use_symlog):
        gamma = self.rddl.discount
        
        # apply discounting of future reward and then optional symlog transform
        def _jax_wrapped_returns(rewards):
            if gamma != 1:
                horizon = rewards.shape[1]
                discount = jnp.power(gamma, jnp.arange(horizon))
                rewards = rewards * discount[jnp.newaxis, ...]
            returns = jnp.sum(rewards, axis=1)
            if use_symlog:
                returns = jnp.sign(returns) * jnp.log1p(jnp.abs(returns))
            return returns
        
        return _jax_wrapped_returns
        
    def _jax_loss(self, rollouts, use_symlog=False): 
        utility_fn = self.utility        
        _jax_wrapped_returns = self._jax_wrapped_returns_fn(use_symlog)
        
        # the loss is the average cumulative reward across all roll-outs
        def _jax_wrapped_plan_loss(key, policy_params, hyperparams,
                                   subs, model_params):
            log = rollouts(key, policy_params, hyperparams, subs, model_params)
            rewards = log['reward']
            returns = _jax_wrapped_returns(rewards)
            utility = utility_fn(returns)
            loss = -utility
            return loss, log
        
        return _jax_wrapped_plan_loss
    
    def _jax_init(self):
        init = self.plan.initializer
        optimizer = self.optimizer
        
        def _jax_wrapped_init_policy(key, hyperparams, subs):
            policy_params = init(key, hyperparams, subs)
            opt_state = optimizer.init(policy_params)
            return policy_params, opt_state
        
        return _jax_wrapped_init_policy
        
    def _jax_update(self, loss):
        optimizer = self.optimizer
        projection = self.plan.projection
        
        # calculate the plan gradient w.r.t. return loss and update optimizer
        # also perform a projection step to satisfy constraints on actions
        def _jax_wrapped_plan_update(key, policy_params, hyperparams,
                                     subs, model_params, opt_state):
            grad_fn = jax.grad(loss, argnums=1, has_aux=True)
            grad, log = grad_fn(key, policy_params, hyperparams, subs, model_params)  
            updates, opt_state = optimizer.update(grad, opt_state) 
            policy_params = optax.apply_updates(policy_params, updates)
            policy_params, converged = projection(policy_params, hyperparams)
            log['grad'] = grad
            log['updates'] = updates
            return policy_params, converged, opt_state, log
        
        return _jax_wrapped_plan_update
            
    def _batched_init_subs(self, subs): 
        rddl = self.rddl
        n_train, n_test = self.batch_size_train, self.batch_size_test
        
        # batched subs
        init_train, init_test = {}, {}
        for (name, value) in subs.items():
            value = np.asarray(value)[np.newaxis, ...]
            train_value = np.repeat(value, repeats=n_train, axis=0)
            train_value = train_value.astype(self.compiled.REAL)
            init_train[name] = train_value
            init_test[name] = np.repeat(value, repeats=n_test, axis=0)
        
        # make sure next-state fluents are also set
        for (state, next_state) in rddl.next_state.items():
            init_train[next_state] = init_train[state]
            init_test[next_state] = init_test[state]
        
        return init_train, init_test
    
    def optimize(self, *args, return_callback: bool=False, **kwargs) -> object:
        ''' Compute an optimal straight-line plan. Returns the parameters
        for the optimized policy.
        
        :param key: JAX PRNG key   
        :param epochs: the maximum number of steps of gradient descent
        :param the maximum number of steps of gradient descent     
        :param train_seconds: total time allocated for gradient descent
        :param plot_step: frequency to plot the plan and save result to disk
        :param model_params: optional model-parameters to override default
        :param policy_hyperparams: hyper-parameters for the policy/plan, such as
        weights for sigmoid wrapping boolean actions
        :param subs: dictionary mapping initial state and non-fluents to 
        their values: if None initializes all variables from the RDDL instance
        :param guess: initial policy parameters: if None will use the initializer
        specified in this instance
        :param verbose: not print (0), print summary (1), print progress (2)
        :param return_callback: whether to return the callback from training
        instead of the parameters
        '''
        * _, callback = self.optimize_generator(*args, **kwargs)
        if return_callback:
            return callback
        else:
            return callback['best_params']
    
    def optimize_generator(self, key: random.PRNGKey,
                           epochs: int=999999,
                           train_seconds: float=120.,
                           plot_step: int=None,
                           model_params: Dict[str, object]=None,
                           policy_hyperparams: Dict[str, object]=None,
                           subs: Dict[str, object]=None,
                           guess: Dict[str, object]=None,
                           verbose: int=2,
                           tqdm_position: int=None) -> Generator[Dict[str, object], None, None]:
        '''Returns a generator for computing an optimal straight-line plan. 
        Generator can be iterated over to lazily optimize the plan, yielding
        a dictionary of intermediate computations.
        
        :param key: JAX PRNG key   
        :param epochs: the maximum number of steps of gradient descent
        :param the maximum number of steps of gradient descent     
        :param train_seconds: total time allocated for gradient descent
        :param plot_step: frequency to plot the plan and save result to disk
        :param model_params: optional model-parameters to override default
        :param policy_hyperparams: hyper-parameters for the policy/plan, such as
        weights for sigmoid wrapping boolean actions
        :param subs: dictionary mapping initial state and non-fluents to 
        their values: if None initializes all variables from the RDDL instance
        :param guess: initial policy parameters: if None will use the initializer
        specified in this instance
        :param verbose: not print (0), print summary (1), print progress (2)
        :param tqdm_position: position of tqdm progress bar (for multiprocessing)
        '''
        verbose = int(verbose)
        start_time = time.time()
        elapsed_outside_loop = 0
        
        # print summary of parameters:
        if verbose >= 1:
            print('==============================================\n'
                  'JAX PLANNER PARAMETER SUMMARY\n'
                  '==============================================')
            self.summarize_hyperparameters()
            print(f'optimize() call hyper-parameters:\n'
                  f'    max_iterations     ={epochs}\n'
                  f'    max_seconds        ={train_seconds}\n'
                  f'    model_params       ={model_params}\n'
                  f'    policy_hyper_params={policy_hyperparams}\n'
                  f'    override_subs_dict ={subs is not None}\n'
                  f'    provide_param_guess={guess is not None}\n' 
                  f'    plot_frequency     ={plot_step}\n')
            
        # compute a batched version of the initial values
        if subs is None:
            subs = self.test_compiled.init_values
        train_subs, test_subs = self._batched_init_subs(subs)
        
        # initialize, model parameters
        if model_params is None:
            model_params = self.compiled.model_params
        model_params_test = self.test_compiled.model_params
        
        # initialize policy parameters
        if guess is None:
            key, subkey = random.split(key)
            policy_params, opt_state = self.initialize(
                subkey, policy_hyperparams, train_subs)
        else:
            policy_params = guess
            opt_state = self.optimizer.init(policy_params)
        best_params, best_loss, best_grad = policy_params, jnp.inf, jnp.inf
        last_iter_improve = 0
        
        # training loop
        iters = range(epochs)
        if verbose >= 2:
            iters = tqdm(iters, total=100, position=tqdm_position)
        
        for it in iters:
            
            # update the parameters of the plan
            key, subkey1, subkey2, subkey3 = random.split(key, num=4)
            policy_params, converged, opt_state, train_log = self.update(
                subkey1, policy_params, policy_hyperparams,
                train_subs, model_params, opt_state)
            if not np.all(converged):
                warnings.warn(
                    f'Projected gradient method for satisfying action concurrency '
                    f'constraints reached the iteration limit: plan is possibly '
                    f'invalid for the current instance.', stacklevel=2)
            
            # evaluate losses
            train_loss, _ = self.train_loss(
                subkey2, policy_params, policy_hyperparams,
                train_subs, model_params)
            test_loss, log = self.test_loss(
                subkey3, policy_params, policy_hyperparams,
                test_subs, model_params_test)
            
            # record the best plan so far
            if test_loss < best_loss:
                best_params, best_loss, best_grad = \
                    policy_params, test_loss, train_log['grad']
                last_iter_improve = it
            
            # save the plan figure
            if plot_step is not None and it % plot_step == 0:
                self._plot_actions(
                    key, policy_params, policy_hyperparams, test_subs, it)
            
            # if the progress bar is used
            elapsed = time.time() - start_time - elapsed_outside_loop
            if verbose >= 2:
                iters.n = int(100 * min(1, max(elapsed / train_seconds, it / epochs)))
                iters.set_description(
                    f'[{tqdm_position}] {it:6} it / {-train_loss:14.4f} train / '
                    f'{-test_loss:14.4f} test / {-best_loss:14.4f} best')
            
            # return a callback
            start_time_outside = time.time()
            yield {
                'iteration': it,
                'train_return':-train_loss,
                'test_return':-test_loss,
                'best_return':-best_loss,
                'params': policy_params,
                'best_params': best_params,
                'last_iteration_improved': last_iter_improve,
                'grad': train_log['grad'],
                'best_grad': best_grad,
                'updates': train_log['updates'],
                'elapsed_time': elapsed,
                'key': key,
                **log
            }
            elapsed_outside_loop += (time.time() - start_time_outside)
            
            # reached time budget
            if elapsed >= train_seconds:
                break
            
            # numerical error
            if not np.isfinite(train_loss):
                break
        
        if verbose >= 2:
            iters.close()
            
        # summarize
        if verbose >= 1:
            grad_norm = jax.tree_map(
                lambda x: np.array(jnp.linalg.norm(x)).item(), best_grad)
            print(f'summary of optimization:\n'
                  f'    time_elapsed  ={elapsed}\n'
                  f'    iterations    ={it}\n'
                  f'    best_objective={-best_loss}\n'
                  f'    grad_norm     ={grad_norm}')
            
    def get_action(self, key: random.PRNGKey,
                   params: Dict,
                   step: int,
                   subs: Dict,
                   policy_hyperparams: Dict[str, object]=None) -> Dict[str, object]:
        '''Returns an action dictionary from the policy or plan with the given
        parameters.
        
        :param key: the JAX PRNG key
        :param params: the trainable parameter PyTree of the policy
        :param step: the time step at which decision is made
        :param policy_hyperparams: hyper-parameters for the policy/plan, such as
        weights for sigmoid wrapping boolean actions
        :param subs: the dict of pvariables
        '''
        actions = self.test_policy(key, params, policy_hyperparams, step, subs)
        grounded_actions = {}
        for (var, action) in actions.items():
            for (ground_var, ground_act) in self.rddl.ground_values(var, action):
                if ground_act != self.noop_actions[ground_var]:
                    grounded_actions[ground_var] = ground_act
        return grounded_actions
    
    def _plot_actions(self, key, params, hyperparams, subs, it):
        rddl = self.rddl
        try:
            import matplotlib.pyplot as plt
        except Exception:
            print('matplotlib is not installed, aborting plot...')
            return
            
        # predict actions from the trained policy or plan
        actions = self.test_rollouts(key, params, hyperparams, subs, {})['action']
            
        # plot the action sequences as color maps
        fig, axs = plt.subplots(nrows=len(actions), constrained_layout=True)
        for (ax, name) in zip(axs, actions):
            action = np.mean(actions[name], axis=0, dtype=float)
            action = np.reshape(action, newshape=(action.shape[0], -1)).T
            if rddl.variable_ranges[name] == 'bool':
                vmin, vmax = 0.0, 1.0
            else:
                vmin, vmax = None, None                
            img = ax.imshow(
                action, vmin=vmin, vmax=vmax, cmap='seismic', aspect='auto')
            ax.set_xlabel('time')
            ax.set_ylabel(name)
            plt.colorbar(img, ax=ax)
            
        # write plot to disk
        plt.savefig(f'plan_{rddl.domainName()}_{rddl.instanceName()}_{it}.pdf',
                    bbox_inches='tight')
        plt.clf()
        plt.close(fig)
    

class JaxRDDLArmijoLineSearchPlanner(JaxRDDLBackpropPlanner):
    '''A class for optimizing an action sequence in the given RDDL MDP using 
    Armijo linear search gradient descent.'''
    
    def __init__(self, *args,
                 optimizer: Callable[..., optax.GradientTransformation]=optax.sgd,
                 optimizer_kwargs: Dict[str, object]={'learning_rate': 1.0},
                 beta: float=0.8,
                 c: float=0.1,
                 lrmax: float=1.0,
                 lrmin: float=1e-5,
                 **kwargs) -> None:
        '''Creates a new gradient-based algorithm for optimizing action sequences
        (plan) in the given RDDL using Armijo line search. All arguments are the
        same as in the parent class, except:
        
        :param beta: reduction factor of learning rate per line search iteration
        :param c: coefficient in Armijo condition
        :param lrmax: initial learning rate for line search
        :param lrmin: minimum possible learning rate (line search halts)
        '''
        self.beta = beta
        self.c = c
        self.lrmax = lrmax
        self.lrmin = lrmin
        super(JaxRDDLArmijoLineSearchPlanner, self).__init__(
            *args,
            optimizer=optimizer,
            optimizer_kwargs=optimizer_kwargs,
            **kwargs)
        
    def summarize_hyperparameters(self):
        super(JaxRDDLArmijoLineSearchPlanner, self).summarize_hyperparameters()
        print(f'linesearch hyper-parameters:\n'
              f'    beta    ={self.beta}\n'
              f'    c       ={self.c}\n'
              f'    lr_range=({self.lrmin}, {self.lrmax})\n')
    
    def _jax_update(self, loss):
        optimizer = self.optimizer
        projection = self.plan.projection
        beta, c, lrmax, lrmin = self.beta, self.c, self.lrmax, self.lrmin
        
        # continue line search if Armijo condition not satisfied and learning
        # rate can be further reduced
        def _jax_wrapped_line_search_armijo_check(val):
            (_, old_f, _, old_norm_g2, _), (_, new_f, lr, _), _, _ = val            
            return jnp.logical_and(
                new_f >= old_f - c * lr * old_norm_g2,
                lr >= lrmin / beta)
            
        def _jax_wrapped_line_search_iteration(val):
            old, new, best, aux = val
            old_x, _, old_g, _, old_state = old
            _, _, lr, iters = new
            _, best_f, _, _ = best
            key, hyperparams, *other = aux
            
            # anneal learning rate and apply a gradient step
            new_lr = beta * lr
            old_state.hyperparams['learning_rate'] = new_lr
            updates, new_state = optimizer.update(old_g, old_state)
            new_x = optax.apply_updates(old_x, updates)
            new_x, _ = projection(new_x, hyperparams)
            
            # evaluate new loss and record best so far
            new_f, _ = loss(key, new_x, hyperparams, *other)
            new = (new_x, new_f, new_lr, iters + 1)
            best = jax.lax.cond(
                new_f < best_f,
                lambda: (new_x, new_f, new_lr, new_state),
                lambda: best
            )
            return old, new, best, aux
            
        def _jax_wrapped_plan_update(key, policy_params, hyperparams,
                                     subs, model_params, opt_state):
            
            # calculate initial loss value, gradient and squared norm
            old_x = policy_params
            loss_and_grad_fn = jax.value_and_grad(loss, argnums=1, has_aux=True)
            (old_f, log), old_g = loss_and_grad_fn(
                key, old_x, hyperparams, subs, model_params)            
            old_norm_g2 = jax.tree_map(lambda x: jnp.sum(jnp.square(x)), old_g)
            old_norm_g2 = jax.tree_util.tree_reduce(jnp.add, old_norm_g2)
            log['grad'] = old_g
            
            # initialize learning rate to maximum
            new_lr = lrmax / beta
            old = (old_x, old_f, old_g, old_norm_g2, opt_state)
            new = (old_x, old_f, new_lr, 0)            
            best = (old_x, jnp.inf, jnp.nan, opt_state)
            aux = (key, hyperparams, subs, model_params)
            
            # do a single line search step with the initial learning rate
            init_val = (old, new, best, aux)            
            init_val = _jax_wrapped_line_search_iteration(init_val)
            
            # continue to anneal the learning rate until Armijo condition holds
            # or the learning rate becomes too small, then use the best parameter
            _, (*_, iters), (best_params, _, best_lr, best_state), _ = \
            jax.lax.while_loop(
                cond_fun=_jax_wrapped_line_search_armijo_check,
                body_fun=_jax_wrapped_line_search_iteration,
                init_val=init_val
            )
            best_state.hyperparams['learning_rate'] = best_lr
            log['updates'] = None
            log['line_search_iters'] = iters
            log['learning_rate'] = best_lr
            return best_params, True, best_state, log
            
        return _jax_wrapped_plan_update

        
class JaxOfflineController(BaseAgent):
    '''A container class for a Jax policy trained offline.'''
    use_tensor_obs = True
    
    def __init__(self, planner: JaxRDDLBackpropPlanner, key: random.PRNGKey,
                 eval_hyperparams: Dict[str, object]=None,
                 params: Dict[str, object]=None,
                 train_on_reset: bool=False,
                 **train_kwargs) -> None:
        '''Creates a new JAX offline control policy that is trained once, then
        deployed later.
        
        :param planner: underlying planning algorithm for optimizing actions
        :param key: the RNG key to seed randomness
        :param eval_hyperparams: policy hyperparameters to apply for evaluation
        or whenever sample_action is called
        :param params: use the specified policy parameters instead of calling
        planner.optimize()
        :param train_on_reset: retrain policy parameters on every episode reset
        :param **train_kwargs: any keyword arguments to be passed to the planner
        for optimization
        '''
        self.planner = planner
        self.key = key
        self.eval_hyperparams = eval_hyperparams
        self.train_on_reset = train_on_reset
        self.train_kwargs = train_kwargs        
        self.params_given = params is not None
        
        self.step = 0
        if not self.train_on_reset and not self.params_given:
            params = self.planner.optimize(key=self.key, **self.train_kwargs) 
        self.params = params  
        
    def sample_action(self, state):
        self.key, subkey = random.split(self.key)
        actions = self.planner.get_action(
            subkey, self.params, self.step, state, self.eval_hyperparams)
        self.step += 1
        return actions
        
    def reset(self):
        self.step = 0
        if self.train_on_reset and not self.params_given:
            self.params = self.planner.optimize(key=self.key, **self.train_kwargs)


class JaxOnlineController(BaseAgent):
    '''A container class for a Jax controller continuously updated using state 
    feedback.'''
    use_tensor_obs = True
    
    def __init__(self, planner: JaxRDDLBackpropPlanner, key: random.PRNGKey,
                 eval_hyperparams: Dict=None, warm_start: bool=True,
                 **train_kwargs) -> None:
        '''Creates a new JAX control policy that is trained online in a closed-
        loop fashion.
        
        :param planner: underlying planning algorithm for optimizing actions
        :param key: the RNG key to seed randomness
        :param eval_hyperparams: policy hyperparameters to apply for evaluation
        or whenever sample_action is called
        :param **train_kwargs: any keyword arguments to be passed to the planner
        for optimization
        '''
        self.planner = planner
        self.key = key
        self.eval_hyperparams = eval_hyperparams
        self.warm_start = warm_start
        self.train_kwargs = train_kwargs
        self.reset()
     
    def sample_action(self, state):
        planner = self.planner
        params = planner.optimize(
            key=self.key,
            guess=self.guess,
            subs=state,
            **self.train_kwargs)
        self.key, subkey = random.split(self.key)
        actions = planner.get_action(subkey, params, 0, state, self.eval_hyperparams)
        if self.warm_start:
            self.guess = planner.plan.guess_next_epoch(params)
        return actions
        
    def reset(self):
        self.guess = None
    



import datetime


class Logger:
    '''Provides functionality for writing messages to a log file.'''
    
    def __init__(self, filename: str) -> None:
        self.filename = filename
    
    def clear(self) -> None:
        fp = open(self.filename, 'w')
        fp.write('')
        fp.close()
    
    def log(self, msg: str) -> None:
        fp = open(self.filename, 'a')
        timestamp = str(datetime.datetime.now())
        fp.write(f'{timestamp}: {msg}\n')
        fp.close()
    
        
class SimLogger:
    '''Provides functionality for writing simulation data to a log file.'''
    
    def __init__(self, filename: str, write_freq: int=1000) -> None:
        self.filename = filename
        self.write_freq = write_freq

    def clear(self, overwrite: bool = True) -> None:
        if overwrite:
            fp = open(self.filename, 'w')
        else:
            fp = open(self.filename, 'a')
        fp.write('')
        fp.close()
        self.data = []
        self.write_head = True
        self.iteration = 0

    def _write_data(self):
        fp = open(self.filename, 'a')
        data = '\n'.join(self.data) + '\n'
        fp.write(data)
        fp.close()
        self.data = [] 
    
    def log(self, obs, action, reward, done, step) -> None:
        if self.write_head:
            header = 'timestamp,iteration,epoch' 
            header += ',' + ','.join(obs.keys())
            header += ',' + ','.join(action.keys())
            header += ',' + 'reward'
            header += ',' + 'done'
            self.data.append(header)
            self.write_head = False
        row = str(datetime.datetime.now())
        row += ',' + str(self.iteration)
        row += ',' + str(step)
        row += ',' + ','.join(map(str, obs.values()))
        row += ',' + ','.join(map(str, action.values()))
        row += ',' + str(reward)
        row += ',' + str(done)
        self.data.append(row)
        self.iteration += 1
        if len(self.data) >= self.write_freq:
            self._write_data()

    def log_free(self, text) -> None:
        self.data.append(text)

    def close(self):
        if self.data:
            self._write_data()



class RDDLGenerator(object):
    def __init__(self, RDDL_AST):
        self.buffer = ""
        self.AST = RDDL_AST

    def GenerateRDDL(self):
        d = self.AST.domain
        self.buffer += self.__print_block_header(d.name)
        self.buffer += '\n'

        # requirements
        if d.requirements:
            self.buffer += self.__print_requirements(d.requirements)
        self.buffer += '\n\n'

        # types
        if d.types:
            self.buffer += self.__print_types([t[0] for t in d.types])
        self.buffer += '\n\n'

        # pvars
        self.buffer += self.__print_pvariables(d)
        self.buffer += '\n\n'

        # cpfs
        self.buffer += self.__print_cpfs(d)
        self.buffer += '\n\n'

        # reward
        self.buffer += self.__print_reward(d.reward)
        self.buffer += '\n\n'

        # constraints
        self.buffer += self.__print_constraints(d)

        self.buffer += '};\n'

        return self.buffer

    def __tabs(self, level):
        t = ''
        for i in range(level):
            t += '\t'
        return t

    def __print_block_header(self, name):
        return 'domain {} '.format(name) + '{'

    def __print_requirements(self, reqs):
        s = self.__tabs(1) + 'requirements = {\n'
        for req in reqs:
            s += self.__tabs(2) + req + ',\n'
        s = s[:-2]
        s += '\n' + self.__tabs(1) + '};'
        return s

    def __print_types(self, types):
        s  = self.__tabs(1) + 'types {\n'
        for type in types:
            s += self.__tabs(2) + type + ': object;\n'
        s += self.__tabs(1) + '};'
        return s

    def __print_pvariables(self, d):
        s = self.__tabs(1) + 'pvariables {\n'
        s += self.__print_typed_pvariables('non-fluent', d.non_fluents)
        s += '\n'
        s += self.__print_typed_pvariables('state-fluent', d.state_fluents)
        s += '\n'
        s += self.__print_typed_pvariables('derived-fluent', d.derived_fluents)
        s += '\n'
        s += self.__print_typed_pvariables('interm-fluent', d.intermediate_fluents)
        s += '\n'
        s += self.__print_typed_pvariables('action-fluent', d.action_fluents)
        s += self.__tabs(1) + '};'
        return s

    def __print_typed_pvariables(self, pvar_type, pvariables):
        s = ""
        for name, pvar in pvariables.items():
            s += self.__tabs(2)
            params = ''
            if pvar.arity > 0:
                params = '(' + ', '.join(pvar.param_types) + ')'
            if pvar.is_derived_fluent() or pvar.is_intermediate_fluent():
                if pvar.level:
                    s += pvar.name + params + ' : {' + pvar_type + ', ' + pvar.range + ', ' + 'level=' + str(pvar.level) + '}\n'
                else:
                    s += pvar.name + params + ' : {' + pvar_type + ', ' + pvar.range + '}\n'
            else:
                s += pvar.name + params + ' : {' + pvar_type + ', ' + pvar.range + ', ' + 'default=' + str(pvar.default) + '}\n'
        return s

    def __print_cpfs(self, d):
        s = self.__tabs(1) + "cpfs {\n"
        s += self.__print_typed_cpfs(d.state_cpfs)
        s += '\n'
        s += self.__print_typed_cpfs(d.intermediate_cpfs)
        s += '\n'
        s += self.__print_typed_cpfs(d.derived_cpfs)
        s += self.__tabs(1) + '};'
        return s

    def __print_typed_cpfs(self, cpfs):
        temp = ''
        for cpf in cpfs:
            name = cpf.pvar[1][0]
            args = '' if cpf.pvar[1][1] is None else '(' + ', '.join(cpf.pvar[1][1]) + ')'
            temp += self.__tabs(2) + name + args +'=\n'
            temp += self.__tabs(4) + self.__scan_expr_tree(cpf.expr)
            temp += '\n'
        return temp

    def __scan_expr_tree(self, expr):#, temp):
        temp = ''
        if expr.etype[0] == 'constant':
            return str(expr.args)
        if expr.etype[0] == 'pvar':
            return self.__format_pvar_args(expr.args)
        if expr.etype[0] == 'relational':
            arg_list = []
            for child in expr.args:
                arg_list.append(self.__scan_expr_tree(child))
            op = expr.etype[1]
            return temp + '(' + arg_list[0] + op + arg_list[1] + ')'
        if expr.etype[0] == 'boolean':
            arg_list = []
            for child in expr.args:
                arg_list.append(self.__scan_expr_tree(child))
            op = expr.etype[1]
            if len(arg_list) == 1:
                return temp + op + arg_list[0]
            else:
                return temp + arg_list[0] + op + arg_list[1]
        if expr.etype[0] == 'control':
            # if
            if_txt = self.__scan_expr_tree(expr.args[0])
            # then
            then_txt = self.__scan_expr_tree(expr.args[1])
            # else
            else_txt = self.__scan_expr_tree(expr.args[2])
            return temp + 'if (' + if_txt + ')\n\tthen ' + then_txt + '\n\telse ' + else_txt
        if expr.etype[0] == 'arithmetic':
            arg_list = []
            for child in expr.args:
                arg_list.append(self.__scan_expr_tree(child))
            op = expr.etype[1]
            if op == '+' or op == '-':
                return temp + op.join(arg_list)
            else:
                return temp + arg_list[0] + op + '(' + arg_list[1] + ')'
        if expr.etype[0] == 'randomvar':
            arg_list = []
            for child in expr.args:
                arg_list.append(self.__scan_expr_tree(child))
            return temp + expr.etype[1] + '(' + ', '.join(arg_list) + ')'
        if expr.etype[0] == 'func':
            temp = expr.etype[1] + '[' + temp
            arg_list = []
            for child in expr.args:
                arg_list.append(self.__scan_expr_tree(child))
            temp = temp + ', '.join(arg_list) + ']'
            return temp
        if expr.etype[0] == 'aggregation':
            inner_expr = self.__scan_expr_tree(expr.args[-1])
            vars = []
            for arg in expr.args[:-1]:
                vars.append(' : '.join(arg[1]))
            temp += '(' + expr.etype[1] + '_{' + ' , '.join(vars) + '} [' + inner_expr + '])'
            return temp
        else:
            for child in expr.args:
                temp += self.__scan_expr_tree(child)
            return temp

    def __format_pvar_args(self, pvar):
        name = pvar[0]
        args = '' if pvar[1] is None else '('+', '.join(pvar[1])+')'
        return name+args

    def __print_constraints(self, d):
        s = ''
        if d.preconds:
            s += self.__print_typed_constraints('action-preconditions', d.preconds)
            s += '\n\n'
        if d.invariants:
            s += self.__print_typed_constraints('state-invariants', d.invariants)
            s += '\n\n'
        if d.constraints:
            s += self.__print_typed_constraints('state-action-constraints', d.constraints)
            s += '\n\n'
        return s

    def __print_typed_constraints(self, constraint_type, constraints):
        const_str = self.__tabs(1) + constraint_type + ' {\n'
        for c in constraints:
            const_str += self.__tabs(2) + self.__scan_expr_tree(c) + ';\n'
        const_str += self.__tabs(1) + '};'
        return const_str

    def __print_reward(self, reward):
        s = self.__tabs(1) + 'reward = '
        s += self.__scan_expr_tree(reward)
        return s


from typing import Union

import sympy as sp
from sympy.logic import boolalg


def get_bernoulli_node_id(
    expr_or_var: Union[sp.Symbol, boolalg.BooleanAtom]) -> int:
    """Returns the node ID associated with the Bernoulli param."""
    # Note: the pattern is f'{dist}_params_{len(args)}_{proba}'.
    if isinstance(expr_or_var, boolalg.BooleanAtom):
        return 1 if expr_or_var else 0
    name = str(expr_or_var)
    dist, _, len_args, param = name.split('_')
    assert dist == 'Bernoulli', (
        "The leaf node must be a Bernoulli node."
    )
    assert len_args == '1', (
        "The length of the arguments must be 1."
    )
    assert param.isdigit(), (
        "The parameter must be a digit corresponding to"
        "the node ID of the Bernoulli parameter."
    )
    return int(param)



import jax
import time
from typing import Dict

from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLActionPreconditionNotSatisfiedError
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLInvalidExpressionError
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLStateInvariantNotSatisfiedError

from pyRDDLGym.Core.Compiler.RDDLLiftedModel import RDDLLiftedModel
from pyRDDLGym.Core.Jax.JaxRDDLCompiler import JaxRDDLCompiler
from pyRDDLGym.Core.Simulator.RDDLSimulator import RDDLSimulator

from pyRDDLGym.Core.Debug.Logger import Logger
from pyRDDLGym.Core.Parser.expr import Value

Args = Dict[str, Value]


class JaxRDDLSimulator(RDDLSimulator):
        
    def __init__(self, rddl: RDDLLiftedModel,
                 key: jax.random.PRNGKey=None,
                 raise_error: bool=True,
                 logger: Logger=None,
                 keep_tensors: bool=False,
                 **compiler_args) -> None:
        '''Creates a new simulator for the given RDDL model with Jax as a backend.
        
        :param rddl: the RDDL model
        :param key: the Jax PRNG key for sampling random variables
        :param raise_error: whether errors are raised as they are encountered in
        the Jax program: unlike the numpy sim, errors cannot be caught in the 
        middle of evaluating a Jax expression; instead they are accumulated and
        returned to the user upon complete evaluation of the expression
        :param logger: to log information about compilation to file
        :param keep_tensors: whether the sampler takes actions and
        returns state in numpy array form
        :param **compiler_args: keyword arguments to pass to the Jax compiler
        '''
        if key is None:
            key = jax.random.PRNGKey(round(time.time() * 1000))
        self.key = key
        self.raise_error = raise_error
        self.compiler_args = compiler_args
        
        # generate direct sampling with default numpy RNG and operations
        super(JaxRDDLSimulator, self).__init__(
            rddl, logger=logger, keep_tensors=keep_tensors)
    
    def seed(self, seed: int) -> None:
        super(JaxRDDLSimulator, self).seed(seed)
        self.key = jax.random.PRNGKey(seed)
        
    def _compile(self):
        rddl = self.rddl
        
        # compilation
        if self.logger is not None:
            self.logger.clear()
        compiled = JaxRDDLCompiler(rddl, logger=self.logger, **self.compiler_args)
        compiled.compile(log_jax_expr=True)
        self.init_values = compiled.init_values
        self.levels = compiled.levels
        self.traced = compiled.traced
        
        self.invariants = jax.tree_map(jax.jit, compiled.invariants)
        self.preconds = jax.tree_map(jax.jit, compiled.preconditions)
        self.terminals = jax.tree_map(jax.jit, compiled.termination)
        self.reward = jax.jit(compiled.reward)
        jax_cpfs = jax.tree_map(jax.jit, compiled.cpfs)
        self.model_params = compiled.model_params
        
        # level analysis
        self.cpfs = []  
        for cpfs in self.levels.values():
            for cpf in cpfs:
                expr = jax_cpfs[cpf]
                prange = rddl.variable_ranges[cpf]
                dtype = compiled.JAX_TYPES.get(prange, compiled.INT)
                self.cpfs.append((cpf, expr, dtype))
        
        # initialize all fluent and non-fluent values    
        self.subs = self.init_values.copy() 
        self.state = None 
        self.noop_actions = {var: values 
                             for (var, values) in self.init_values.items() 
                             if rddl.variable_types[var] == 'action-fluent'}
        self._pomdp = bool(rddl.observ)
        
        # cached for performance
        self.invariant_names = [f'Invariant {i}' for i in range(len(rddl.invariants))]        
        self.precond_names = [f'Precondition {i}' for i in range(len(rddl.preconditions))]
        self.terminal_names = [f'Termination {i}' for i in range(len(rddl.terminals))]
        
        self.grounded_actionsranges = rddl.groundactionsranges()
        self.grounded_noop_actions = rddl.ground_values_from_dict(self.noop_actions)
        
    def handle_error_code(self, error, msg) -> None:
        if self.raise_error:
            errors = JaxRDDLCompiler.get_error_messages(error)
            if errors:
                message = f'Internal error in evaluation of {msg}:\n'
                errors = '\n'.join(f'{i + 1}. {s}' for (i, s) in enumerate(errors))
                raise RDDLInvalidExpressionError(message + errors)
    
    def check_state_invariants(self, silent: bool=False) -> bool:
        '''Throws an exception if the state invariants are not satisfied.'''
        for (i, invariant) in enumerate(self.invariants):
            loc = self.invariant_names[i]
            sample, self.key, error = invariant(
                self.subs, self.model_params, self.key)
            self.handle_error_code(error, loc)            
            if not bool(sample):
                if not silent:
                    raise RDDLStateInvariantNotSatisfiedError(
                        f'{loc} is not satisfied.')
                return False
        return True
    
    def check_action_preconditions(self, actions: Args, silent: bool=False) -> bool:
        '''Throws an exception if the action preconditions are not satisfied.'''
        actions = self._process_actions(actions)
        subs = self.subs
        subs.update(actions)
        
        for (i, precond) in enumerate(self.preconds):
            loc = self.precond_names[i]
            sample, self.key, error = precond(subs, self.model_params, self.key)
            self.handle_error_code(error, loc)            
            if not bool(sample):
                if not silent:
                    raise RDDLActionPreconditionNotSatisfiedError(
                        f'{loc} is not satisfied for actions {actions}.')
                return False
        return True
    
    def check_terminal_states(self) -> bool:
        '''return True if a terminal state has been reached.'''
        for (i, terminal) in enumerate(self.terminals):
            loc = self.terminal_names[i]
            sample, self.key, error = terminal(
                self.subs, self.model_params, self.key)
            self.handle_error_code(error, loc)
            if bool(sample):
                return True
        return False
    
    def sample_reward(self) -> float:
        '''Samples the current reward given the current state and action.'''
        reward, self.key, error = self.reward(
            self.subs, self.model_params, self.key)
        self.handle_error_code(error, 'reward function')
        return float(reward)
    
    def step(self, actions: Args) -> Args:
        '''Samples and returns the next state from the cpfs.
        
        :param actions: a dict mapping current action fluents to their values
        '''
        rddl = self.rddl
        keep_tensors = self.keep_tensors
        actions = self._process_actions(actions)
        subs = self.subs
        subs.update(actions)
        
        # compute CPFs in topological order
        for (cpf, expr, _) in self.cpfs:
            subs[cpf], self.key, error = expr(subs, self.model_params, self.key)
            self.handle_error_code(error, f'CPF <{cpf}>')            
                
        # sample reward
        reward = self.sample_reward()
        
        # update state
        self.state = {}
        for (state, next_state) in rddl.next_state.items():
            subs[state] = subs[next_state]
            if keep_tensors:
                self.state[state] = subs[state]
            else:
                self.state.update(rddl.ground_values(state, subs[state]))
        
        # update observation
        if self._pomdp: 
            obs = {}
            for var in rddl.observ:
                if keep_tensors:
                    obs[var] = subs[var]
                else:
                    obs.update(rddl.ground_values(var, subs[var]))
        else:
            obs = self.state
        
        done = self.check_terminal_states()        
        return obs, reward, done
        



from abc import ABCMeta, abstractmethod
import numpy as np
import random
import gym
from typing import Dict

from pyRDDLGym.Core.Env.RDDLEnv import RDDLEnv


class BaseAgent(metaclass=ABCMeta):
    '''Base class for policies.'''
    use_tensor_obs = False  # uses internal tensor representation of state
    
    @abstractmethod
    def sample_action(self, state: object) -> object:
        '''Samples an action from the current policy evaluated at the given state.
        
        :param state: the current state
        '''
        pass
    
    def reset(self) -> None:
        '''Resets the policy and prepares it for the next episode.'''
        pass
    
    def evaluate(self, env: RDDLEnv, episodes: int=1, 
                 verbose: bool=False, render: bool=False, 
                 seed: int=None) -> Dict[str, float]:
        '''Evaluates the current agent on the specified environment by simulating
        roll-outs. Returns a dictionary of summary statistics of the returns
        accumulated on the roll-outs.
        
        :param env: the environment
        :param episodes: how many episodes (trials) to perform
        :param verbose: whether to print the transition information to console
        at each step of the simulation
        :param render: visualize the domain using the env internal visualizer
        :param seed: optional RNG seed for the environment
        '''
        gamma = env.discount
        
        history = np.zeros((episodes,))
        for episode in range(episodes):
            
            # restart episode
            total_reward, cuml_gamma = 0.0, 1.0
            self.reset()
            if env.new_gym_api:
                state, _ = env.reset(seed=seed)
            else:
                state = env.reset(seed=seed)
            
            # simulate to end of horizon
            for step in range(env.horizon):
                if render:
                    env.render()
                
                # take a step in the environment
                if self.use_tensor_obs:
                    policy_input = env.sampler.subs
                else:
                    policy_input = state
                action = self.sample_action(policy_input)   
                if env.new_gym_api: 
                    next_state, reward, done, _, _ = env.step(action)
                else:
                    next_state, reward, done, _ = env.step(action)
                total_reward += reward * cuml_gamma
                cuml_gamma *= gamma
                
                # printing
                if verbose: 
                    print(f'step       = {step}\n'
                          f'state      = {state}\n'
                          f'action     = {action}\n'
                          f'next state = {next_state}\n'
                          f'reward     = {reward}\n')
                state = next_state
                if done:
                    break
            
            if verbose:
                print(f'episode {episode + 1} ended with return {total_reward}')
            history[episode] = total_reward
        
        # summary statistics
        return {
            'mean': np.mean(history),
            'median': np.median(history),
            'min': np.min(history),
            'max': np.max(history),
            'std': np.std(history)
        }


class RandomAgent(BaseAgent):
    '''Uniformly pseudo-random policy.'''

    def __init__(self, action_space, num_actions=1, seed=None):
        '''Creates a new uniformly pseudo-random policy.
        
        :param action_space: the set of actions from which to sample uniformly
        :param num_actions: the number of samples to produce
        '''
        self.action_space = action_space
        self.num_actions = num_actions
        self.rng = random.Random(seed)
        if seed is not None:
            self.action_space.seed(seed)

    def sample_action(self, state=None):
        s = self.action_space.sample()
        action = {}
        selected_actions = self.rng.sample(list(s), self.num_actions)
        for sample in selected_actions:
            if isinstance(self.action_space[sample], gym.spaces.Box):
                action[sample] = s[sample][0].item()
            elif isinstance(self.action_space[sample], gym.spaces.Discrete):
                action[sample] = s[sample]
                # if str(self.action_space[sample]) == 'Discrete(2)':
                #     action[sample] = bool(s[sample])
                # else:
                #     action[sample] = s[sample]
        return action


class NoOpAgent(BaseAgent):
    '''No-op policy.'''
    
    def __init__(self, action_space, num_actions=0):
        '''Creates a new no-op policy.
            
        :param action_space: the set of actions (currently unused)
        :param num_actions: the number of samples to produce (currently unused)
        '''
        self.action_space = action_space
        self.num_actions = num_actions

    def sample_action(self, state=None):
        action = {}
        return action




from typing import Dict

import numpy as np
from pyRDDLGym.Examples.InstanceGenerator import InstanceGenerator
import tensorflow as tf
import tensorflow_probability as tfp

tfd = tfp.distributions


class RecSimInstanceGenerator(InstanceGenerator):

  def get_env_path(self) -> str:
    return 'RecSim'

  def get_domain_name(self) -> str:
    return 'recsim_ecosystem_welfare'

  def init_random_provider_clusters(
      self,
      provider_disp: float,
      provider_fan_out: int,
      num_provider_clusters: int,
      num_topics: int,
  ) -> np.ndarray:
    """Initializes a set of providers over different topics for mixtures."""
    provider_cluster_seed = np.float32(
        (
            (provider_disp**0.5)
            * np.random.randn(num_provider_clusters, num_topics)
        )
    )
    providers = None
    for provider_cluster_center in range(num_provider_clusters):
      provider_cluster = np.float32(
          ((0.5**0.5) * np.random.randn(provider_fan_out, num_topics))
          + provider_cluster_seed[provider_cluster_center, :]
      )
      if providers is None:
        providers = provider_cluster
      else:
        providers = np.vstack((providers, provider_cluster))
    return np.array(providers)

  def init_random_user_points(self, provider_means, num_users, user_stddev):
    provider_logits = -tf.math.log(1.0 + tf.norm(provider_means, axis=1))
    batch_provider_logits = tf.broadcast_to(
        tf.expand_dims(provider_logits, axis=0),
        [num_users] + provider_logits.shape,
    )

    dist = tfd.MixtureSameFamily(
        mixture_distribution=tfd.Categorical(logits=batch_provider_logits),
        components_distribution=tfd.MultivariateNormalDiag(
            loc=provider_means, scale_identity_multiplier=user_stddev
        ),
    )
    return dist.sample()

  def sample_instance(self, params: Dict[str, object]) -> Dict[str, object]:
    provider_disp = params['provider_dispersion']
    provider_fan_out = params['provider_fan_out']
    num_provider_clusters = params['num_provider_clusters']
    num_topics = 2  # These problems are always 2-dimensional.
    num_users = params['num_users']
    num_docs = num_provider_clusters * params['docs_per_cluster']
    user_stddev = params['user_stddev']
    provider_clusters = self.init_random_provider_clusters(
        provider_disp, provider_fan_out, num_provider_clusters, num_topics
    )
    user_locs = self.init_random_user_points(
        provider_clusters, num_users, user_stddev
    )

    # Begin instance generation.
    # Objects.
    providers = [f'p{i+1}' for i in range(num_provider_clusters)]
    users = [f'c{i+1}' for i in range(num_users)]
    features = ['f1', 'f2']
    items = [f'i{i+1}' for i in range(num_docs)]
    # Non-fluents.
    nonfluents = {}
    for i, user in enumerate(users):
      nonfluents[f'CONSUMER-AFFINITY({user}, f1)'] = user_locs[i][0].numpy()
      nonfluents[f'CONSUMER-AFFINITY({user}, f2)'] = user_locs[i][1].numpy()
    for i, provider in enumerate(providers):
      nonfluents[f'PROVIDER-COMPETENCE({provider}, f1)'] = provider_clusters[i][
          0
      ]
      nonfluents[f'PROVIDER-COMPETENCE({provider}, f2)'] = provider_clusters[i][
          1
      ]
    providers = [
        'pn'
    ] + providers  # Add dummy provider for total ordering purposes.
    for i, provider in enumerate(providers[:-1]):
      nonfluents[f'NEXT-PROVIDER({provider}, {providers[i+1]})'] = True
    for i, provider in enumerate(providers[:-1]):
      for next_provider in providers[i + 1 :]:
        nonfluents[f'LESS({provider}, {next_provider})'] = True
    nonfluents['MAX-AFFINITY'] = 50.0
    # Initial states.
    states = {'provider-satisfaction(pn)': 0.0}
    return {
        'objects': {
            'feature': features,
            'item': items,
            'consumer': users,
            'provider': providers,
        },
        'non-fluents': nonfluents,
        'init-states': states,
        'horizon': params['horizon'],
        'discount': params['discount'],
        'max-nondef-actions': 'pos-inf',
    }





import argparse

from pyRDDLGym.Visualizer.visualize_dbn import RDDL2Graph


def test_dbn_visualization(args: argparse.Namespace):
    domains = [
        'cartpole discrete',
        'cartpole continuous',
        'marsrover', 
        # 'mountaincar',    # Takes too long to create DBN graph for this
        'powergen continuous',
        'powergen discrete',
        'racecar',
        'uavcontinuous', 
        'uavdiscrete',
        'uavmixed',
        'wildfire',
        'supplychain', 
    ]
    
    for domain in domains:
        print(f"Creating DBN graph for {domain}...")
        r2g = RDDL2Graph(
            domain=domain,
            instance=0,
            directed=True,
            strict_grouping=True,
            simulation=args.simulation,
        )
        r2g.save_dbn(file_name=domain)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('--simulation', action='store_true')
    args = parser.parse_args()

    test_dbn_visualization(args)



import os

from pyRDDLGym import RDDLEnv
from pyRDDLGym.Core.Env.RDDLEnvSeeder import RDDLEnvSeeder
from pyRDDLGym.Core.Policies.Agents import NoOpAgent as RDDLAgent


def Test_instance(dom_path, inst_path, episodes=1):
    myEnv = RDDLEnv.RDDLEnv(domain=dom_path,
                            instance=inst_path,
                            enforce_action_constraints=False,
                            debug=False,
                            seeds=[42, 43])
    agent = RDDLAgent(action_space=myEnv.action_space,
                      num_actions=myEnv.numConcurrentActions)

    for episode in range(episodes):
        total_reward = 0
        state = myEnv.reset()
        # seeder = RDDLEnvSeeder(seed_list=[42, 43])
        # state = myEnv.reset(seed=seeder.Next())
        for step in range(myEnv.horizon):
            myEnv.render()
            action = agent.sample_action()
            next_state, reward, done, info = myEnv.step(action)
            total_reward += reward
            print()
            print(f'step       = {step}')
            print(f'state      = {state}')
            print(f'action     = {action}')
            print(f'next state = {next_state}')
            print(f'reward     = {reward}')
            state = next_state
            if done:
                break
        print(f'episode ended with reward {total_reward}')
    myEnv.close()


def main(path):
    print(path)

    dom_path = os.path.join(path, 'domain.rddl')
    inst_path = os.path.join(path, 'instance.rddl')
    print(dom_path)
    print(inst_path)

    Test_instance(dom_path, inst_path, 3)


if __name__ == "__main__":

    path = os.path.dirname(os.path.abspath(__file__))
    path = os.path.join(path, 'SeedTest')

    main(path)



import matplotlib
matplotlib.use('agg')
import matplotlib.pyplot as plt
import numpy as np
from PIL import Image
import pprint

from pyRDDLGym.Core.Compiler.RDDLModel import PlanningModel
from pyRDDLGym import Visualizer
from pyRDDLGym.Visualizer.StateViz import StateViz


class TextVisualizer(StateViz):

    def __init__(self, model: PlanningModel,
                 figure_size=[5, 10],
                 dpi=100,
                 fontsize=10,
                 display=False) -> None:
        self._model = model
        self._states = model.groundstates()
        self._nonfluents = model.groundnonfluents()
        self._objects = model.objects
        self._figure_size = figure_size
        self._display = display
        self._dpi = dpi
        self._fontsize = fontsize
        self._interval = 10
        self._asset_path = "/".join(Visualizer.__file__.split("/")[:-1])
        self._nonfluents_layout = None
        self._states_layout = None
        self._fig, self._ax = None, None
        self._data = None
        self._img = None

        # if display == True:
        #     self.pygame_thread = threading.Thread(target=self.init_display)
        #     self.pygame_thread.start()

    def build_nonfluents_layout(self):
        return self._nonfluents
    
    def build_states_layout(self, states):
        return states
    
    def init_canvas(self, figure_size, dpi):
        fig = plt.figure(figsize=figure_size, dpi=dpi)
        ax = plt.gca()
        plt.xlim(0, figure_size[0] * self._interval)
        plt.ylim(0, figure_size[1] * self._interval)
        plt.axis('scaled')
        plt.axis('off')
        return fig, ax
        
    def convert2img(self, fig, ax):
        
        ax.set_position((0, 0, 1, 1))
        fig.canvas.draw()

        data = np.frombuffer(fig.canvas.tostring_rgb(), dtype=np.uint8)
        data = data.reshape(fig.canvas.get_width_height()[::-1] + (3,))

        img = Image.fromarray(data)

        self._data = data
        self._img = img

        return img

    def render(self, state):
        self.states = state

        self._fig, self._ax = self.init_canvas(self._figure_size, self._dpi)
        
        state_layout = self.build_states_layout(state)
        text_layout = {'state': state_layout}

        text_str = pprint.pformat(text_layout)[1:-1]
        self._ax.text(self._interval * 0.5,
                      self._figure_size[1] * self._interval * 0.95,
                      text_str,
                      horizontalalignment='left',
                      verticalalignment='top',
                      wrap=True, fontsize=self._fontsize)
        
        img = self.convert2img(self._fig, self._ax)

        self._ax.cla()
        plt.close()
        # plt.show()

        return img
    
